from pathlib import Path
from nonebot import on_command
from nonebot.internal.adapter import Message
from nonebot.params import CommandArg
from nonebot_plugin_alconna import UniMessage


osu_help = on_command("osuhelp", aliases={"OSUBot", "osubot"}, priority=11, block=True)

with open(Path(__file__).parent.parent / "osufile" / "help.png", "rb") as f:
    img1 = f.read()
with open(Path(__file__).parent.parent / "osufile" / "detail.png", "rb") as f:
    img2 = f.read()


@osu_help.handle()
async def _help(arg: Message = CommandArg()):
    arg = arg.extract_plain_text().strip()
    if arg == "detail":
        await UniMessage.image(raw=img2).finish(reply_to=True)
    await UniMessage.image(raw=img1).finish(reply_to=True)



import re
from io import BytesIO, TextIOWrapper
from typing import Union, Optional
from pathlib import Path
from nonebot.log import logger
from .api import safe_async_get
from .network import auto_retry
from .network.first_response import get_first_response
from .schema import Badge

osufile = Path(__file__).parent / "osufile"
map_path = Path() / "data" / "osu" / "map"
user_cache_path = Path() / "data" / "osu" / "user"
badge_cache_path = Path() / "data" / "osu" / "badge"
api_ls = [
    "https://api.chimu.moe/v1/download/",
    "https://api.osu.direct/d/",
    "https://txy1.sayobot.cn/beatmaps/download/novideo/",
]

if not map_path.exists():
    map_path.mkdir(parents=True, exist_ok=True)
if not user_cache_path.exists():
    user_cache_path.mkdir(parents=True, exist_ok=True)
if not badge_cache_path.exists():
    badge_cache_path.mkdir(parents=True, exist_ok=True)


async def download_map(setid: int) -> Optional[Path]:
    urls = [i + str(setid) for i in api_ls]
    logger.info(f"开始下载地图: <{setid}>")
    req = await get_first_response(urls)
    filename = f"{setid}.osz"
    filepath = map_path.parent / filename
    with open(filepath, "wb") as f:
        f.write(req.read())
    logger.info(f"地图: <{setid}> 下载完毕")
    return filepath


async def download_tmp_osu(map_id):
    url = f"https://osu.ppy.sh/osu/{map_id}"
    logger.info(f"开始下载谱面: <{map_id}>")
    req = await safe_async_get(url)
    filename = f"{map_id}.osu"
    filepath = map_path / filename
    chunk = req.read()
    with open(filepath, "wb") as f:
        f.write(chunk)
    return filepath


@auto_retry
async def download_osu(set_id, map_id):
    url = [f"https://osu.ppy.sh/osu/{map_id}", f"https://api.osu.direct/osu/{map_id}"]
    logger.info(f"开始下载谱面: <{map_id}>")
    if req := await get_first_response(url):
        filename = f"{map_id}.osu"
        filepath = map_path / str(set_id) / filename
        with open(filepath, "wb") as f:
            f.write(req)
        return filepath
    else:
        raise Exception("下载出错，请稍后再试")


async def get_projectimg(url: str):
    if "avatar-guest.png" in url:
        url = "https://osu.ppy.sh/images/layout/avatar-guest.png"
    req = await safe_async_get(url)
    if not req or req.status_code == 403:
        return osufile / "work" / "mapbg.png"
    data = req.read()
    im = BytesIO(data)
    return im


def re_map(file: Union[bytes, Path]) -> str:
    if isinstance(file, bytes):
        text = TextIOWrapper(BytesIO(file), "utf-8").read()
    else:
        with open(file, "r", encoding="utf-8") as f:
            text = f.read()
    res = re.search(r"\d,\d,\"(.+)\"", text)
    bg = "mapbg.png" if not res else res.group(1).strip()
    return bg


async def make_badge_cache_file(badge: Badge):
    path = badge_cache_path / f"{hash(badge.description)}.png"
    badge_icon = await get_projectimg(badge.image_url)
    with open(path, "wb") as f:
        f.write(badge_icon.getvalue())


# 保存个人信息界面背景
async def save_info_pic(user: str, byt: bytes):
    path = user_cache_path / user
    if not path.exists():
        path.mkdir()
    with open(path / "info.png", "wb") as f:
        f.write(BytesIO(byt).getvalue())



import re
from random import shuffle

from expiringdict import ExpiringDict
from nonebot import on_command
from nonebot_plugin_alconna import UniMessage
from nonebot.internal.matcher import Matcher
from nonebot.typing import T_State
from nonebot.log import logger

from .utils import split_msg
from ..api import get_sayo_map_info, get_recommend, update_recommend

recommend = on_command(
    "推荐", priority=11, block=True, aliases={"recommend", "推荐铺面", "推荐谱面"}
)
recommend_cache = ExpiringDict(1000, 60 * 60 * 12)


async def handle_recommend(state: T_State, matcher: type[Matcher]):
    user = state["user"]
    mode = state["mode"]
    mods = state["mods"]
    if mods == ["4K"]:
        key_count = "4"
    elif mods == ["7K"]:
        key_count = "7"
    else:
        key_count = "4,7"
    if mode == "1" or mode == "2":
        await matcher.finish("很抱歉，该模式暂不支持推荐")
    if not recommend_cache.get(user):
        recommend_cache[user] = set()
        await update_recommend(user)
    recommend_data = await get_recommend(user, mode, key_count)
    if not recommend_data.data.list:
        await matcher.finish("没有可以推荐的图哦，自己多打打喜欢玩的图吧")
    shuffle(recommend_data.data.list)
    for i in recommend_data.data.list:
        if i.id not in recommend_cache[user]:
            recommend_cache[user].add(i.id)
            recommend_map = i
            break
    else:
        await matcher.finish("今天已经没有可以推荐的图啦，明天再来吧")
    bid = int(re.findall("https://osu.ppy.sh/beatmaps/(.*)", recommend_map.mapLink)[0])
    map_info = await get_sayo_map_info(bid, 1)
    sid = map_info.data.sid
    for i in map_info.data.bid_data:
        if i.bid == bid:
            bg = i.bg
            break
    else:
        bg = ""
        logger.debug(f"如果看到这句话请联系作者 有问题的是{bid}, {sid}")
    s = (
        f'推荐的铺面是{recommend_map.mapName} ⭐{round(recommend_map.difficulty, 2)}\n{"".join(recommend_map.mod)}\n'
        f"预计pp为{round(recommend_map.predictPP, 2)}\n提升概率为{round(recommend_map.passPercent*100, 2)}%\n"
        f"{recommend_map.mapLink}\nhttps://kitsu.moe/api/d/{sid}\n"
        f"https://txy1.sayobot.cn/beatmaps/download/novideo/{sid}"
    )
    pic_url = f"https://dl.sayobot.cn/beatmaps/files/{sid}/{bg}"
    return pic_url, s


@recommend.handle(parameterless=[split_msg()])
async def _(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).finish(reply_to=True)
    pic_url, s = await handle_recommend(state, recommend)
    await (UniMessage.image(url=pic_url) + s).finish(reply_to=True)



from io import BytesIO
from typing import Union
from PIL import Image, UnidentifiedImageError

from ..api import osu_api, get_map_bg
from ..file import download_osu, re_map, map_path


async def get_bg(mapid: Union[str, int]) -> Union[str, BytesIO]:
    info = await osu_api("map", map_id=mapid)
    if not info:
        return "未查询到该地图"
    elif isinstance(info, str):
        return info
    setid: int = info["beatmapset_id"]
    set_path = map_path / str(setid)
    if not set_path.exists():
        set_path.mkdir(parents=True, exist_ok=True)
    osu = map_path / str(setid) / f"{mapid}.osu"
    if not osu.exists():
        await download_osu(setid, mapid)
    cover = re_map(osu)
    cover_path = map_path / str(setid) / cover
    if not cover_path.exists():
        if bg := await get_map_bg(mapid, setid, cover):
            with open(cover_path, "wb") as f:
                f.write(bg.getvalue())
    try:
        img = Image.open(cover_path)
    except UnidentifiedImageError:
        cover_path.unlink()
        return "暂时无法下载背景图片＞︿＜"
    byt = BytesIO()
    img.save(byt, "png")
    return byt



from nonebot import on_command
from nonebot.typing import T_State
from nonebot_plugin_alconna import UniMessage
from .utils import split_msg


mu = on_command("mu", priority=11, block=True)


@mu.handle(parameterless=[split_msg()])
async def _mu(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).finish(reply_to=True)
    await UniMessage.text(f"https://osu.ppy.sh/u/{state['user']}").finish(reply_to=True)



from nonebot import on_command
from nonebot.internal.adapter import Event, Message
from nonebot.params import CommandArg
from nonebot_plugin_alconna import UniMessage

from ..database import UserData
from ..utils import NGM


update_mode = on_command("更新模式", priority=11, block=True)


@update_mode.handle()
async def _(event: Event, mode: Message = CommandArg()):
    mode = mode.extract_plain_text().strip()
    user = await UserData.get_or_none(user_id=event.get_user_id())
    if not user:
        await UniMessage.text("该账号尚未绑定，请输入 /bind 用户名 绑定账号").finish(reply_to=True)
    elif not mode:
        await UniMessage.text("请输入需要更新内容的模式").finish(reply_to=True)
    if mode.isdigit() and 0 <= int(mode) < 4:
        await UserData.filter(user_id=event.get_user_id()).update(osu_mode=int(mode))
        msg = f"已将默认模式更改为 {NGM[mode]}"
    else:
        msg = "请输入正确的模式 0-3"
    await UniMessage.text(msg).finish(reply_to=True)



from pathlib import Path

from nonebot.exception import ParserExit, ActionFailed
from nonebot.internal.adapter import Message
from nonebot.params import ShellCommandArgv, CommandArg
from nonebot.rule import ArgumentParser
from nonebot import on_command, on_shell_command
from nonebot_plugin_alconna import UniMessage
from ..api import get_sayo_map_info

from ..mania import convert_mania_map, Options

parser = ArgumentParser('convert', description='变换mania谱面')
parser.add_argument('--set', type=int, help='要转换的谱面的setid')
parser.add_argument('--map', type=int, help='要转换的谱面的mapid')
parser.add_argument('--fln', action='store_true', help='将谱面转换为反键')
parser.add_argument('--rate', type=float, help='谱面倍速速率')
parser.add_argument('--end_rate', type=float, help='谱面倍速速率的最大值')
parser.add_argument('--step', type=float, help='谱面倍速的step')
parser.add_argument('--od', type=float, help='改变谱面od到指定值')
parser.add_argument('--nsv', action='store_true', help='移除谱面所有sv')
parser.add_argument('--nln', action='store_true', help='移除谱面所有ln')
parser.add_argument('--gap', nargs='?', default='150', type=float, help='指定反键的间距时间，默认150ms')
parser.add_argument('--thres', nargs='?', default='100', type=float, help='指定转反键时ln转换为note的阈值，默认100ms')

convert = on_shell_command("convert", parser=parser, block=True, priority=13)


@convert.handle()
async def _(argv: list[str] = ShellCommandArgv()):
    try:
        args = parser.parse_args(argv)
    except ParserExit as e:
        if e.status == 0:
            await UniMessage.image(path=Path(__file__).parent / 'osufile' / 'convert.jpg').finish(reply_to=True)
        await UniMessage.text(str(e)).finish(reply_to=True)
        return
    options = Options(**vars(args))
    if options.map:
        sayo_map_info = await get_sayo_map_info(options.map, 1)
        options.set = sayo_map_info.data.sid
        options.sayo_info = sayo_map_info
    if not options.set:
        await UniMessage.text('请提供需要转换的谱面setid').finish(reply_to=True)
    if options.nln and options.fln:
        await UniMessage.text('指令矛盾！').finish(reply_to=True)
    osz_file = await convert_mania_map(options)
    if not osz_file:
        await UniMessage.text('未找到该地图，请检查是否搞混了mapID与setID').finish(reply_to=True)
    file_path = osz_file.absolute()
    try:
        with open(file_path, 'rb') as f:
            await UniMessage.file(raw=f.read()).send()
    except ActionFailed:
        await UniMessage.text('上传文件失败，可能是群空间满或没有权限导致的').send(reply_to=True)
    finally:
        try:
            osz_file.unlink()
        except PermissionError:
            ...


change = on_command('倍速', priority=11, block=True)


@change.handle()
async def _(msg: Message = CommandArg()):
    args = msg.extract_plain_text().strip().split()
    argv = ['--map']
    if not args:
        await UniMessage.text('请输入需要倍速的地图mapID').finish(reply_to=True)
    set_id = args[0]
    if not set_id.isdigit():
        await UniMessage.text('请输入正确的mapID').finish(reply_to=True)
    argv.append(set_id)
    if len(args) >= 2:
        argv.append('--rate')
        if '-' in args[1]:
            low, high = args[1].split('-')
            argv.extend([low, '--end_rate', high, '--step', '0.05'])
        else:
            argv.append(args[1])
    else:
        await UniMessage.text('请输入倍速速率').finish(reply_to=True)
    args = parser.parse_args(argv)
    options = Options(**vars(args))
    if options.map:
        sayo_map_info = await get_sayo_map_info(options.map, 1)
        options.set = sayo_map_info.data.sid
        options.sayo_info = sayo_map_info
    osz_path = await convert_mania_map(options)
    if not osz_path:
        await UniMessage.text('未找到该地图，请检查是否搞混了mapID与setID').finish(reply_to=True)
    file_path = osz_path.absolute()
    try:
        with open(file_path, 'rb') as f:
            await UniMessage.file(raw=f.read()).send()
    except ActionFailed:
        await UniMessage.text('上传文件失败，可能是群空间满或没有权限导致的').send(reply_to=True)
    finally:
        try:
            osz_path.unlink()
        except PermissionError:
            ...


generate_full_ln = on_command('反键', priority=11, block=True)


@generate_full_ln.handle()
async def _(msg: Message = CommandArg()):
    args = msg.extract_plain_text().strip().split()
    if not args:
        await UniMessage.text('请输入需要转ln的地图setID').finish(reply_to=True)
    set_id = args[0]
    if not set_id.isdigit():
        await UniMessage.text('请输入正确的setID').finish(reply_to=True)
    argv = ['--set', set_id, '--fln']
    if len(args) >= 2:
        argv.append('--gap')
        argv.append(args[1])
    if len(args) >= 3:
        argv.append('--thres')
        argv.append(args[2])
    args = parser.parse_args(argv)
    options = Options(**vars(args))
    osz_path = await convert_mania_map(options)
    if not osz_path:
        await UniMessage.text('未找到该地图，请检查是否搞混了mapID与setID').finish(reply_to=True)
    file_path = osz_path.absolute()
    try:
        with open(file_path, 'rb') as f:
            await UniMessage.file(raw=f.read()).send()
    except ActionFailed:
        await UniMessage.text('上传文件失败，可能是群空间满或没有权限导致的').send(reply_to=True)
    finally:
        try:
            osz_path.unlink()
        except PermissionError:
            ...



import os
import random

from io import BytesIO
from typing import Optional, Union
from PIL import ImageDraw, UnidentifiedImageError, ImageEnhance, ImageFilter
from matplotlib.figure import Figure

from ..file import get_projectimg, user_cache_path
from ..schema import SeasonalBackgrounds, User
from ..api import get_seasonal_bg, safe_async_get

from .static import *


def image2bytesio(pic: Image):
    byt = BytesIO()
    pic.save(byt, "png")
    return byt


def draw_fillet(img, radii):
    # 画圆（用于分离4个角）
    circle = Image.new("L", (radii * 2, radii * 2), 0)  # 创建一个黑色背景的画布
    draw = ImageDraw.Draw(circle)
    draw.ellipse((0, 0, radii * 2, radii * 2), fill=255)  # 画白色圆形
    # 原图
    img = img.convert("RGBA")
    w, h = img.size
    # 画4个角（将整圆分离为4个部分）
    alpha = Image.new("L", img.size, 255)
    alpha.paste(circle.crop((0, 0, radii, radii)), (0, 0))  # 左上角
    alpha.paste(circle.crop((radii, 0, radii * 2, radii)), (w - radii, 0))  # 右上角
    alpha.paste(
        circle.crop((radii, radii, radii * 2, radii * 2)), (w - radii, h - radii)
    )  # 右下角
    alpha.paste(circle.crop((0, radii, radii, radii * 2)), (0, h - radii))  # 左下角
    # 白色区域透明可见，黑色区域不可见
    img.putalpha(alpha)
    return img


def draw_fillet2(img, radii):
    # 画圆（用于分离4个角）
    circle = Image.new("L", (radii * 2, radii * 2), 0)  # 创建一个黑色背景的画布
    draw = ImageDraw.Draw(circle)
    draw.ellipse((0, 0, radii * 2, radii * 2), fill=255)  # 画白色圆形
    # 原图
    img = img.convert("RGBA")
    w, h = img.size
    # 创建一个透明度为50%的图像
    enhancer = ImageEnhance.Brightness(img)
    img = enhancer.enhance(0.5)
    # 画4个角（将整圆分离为4个部分）
    alpha = Image.new("L", img.size, 255)
    alpha.paste(circle.crop((0, 0, radii, radii)), (0, 0))  # 左上角
    alpha.paste(circle.crop((radii, 0, radii * 2, radii)), (w - radii, 0))  # 右上角
    alpha.paste(
        circle.crop((radii, radii, radii * 2, radii * 2)), (w - radii, h - radii)
    )  # 右下角
    alpha.paste(circle.crop((0, radii, radii, radii * 2)), (0, h - radii))  # 左下角
    # 高斯模糊效果
    img = img.filter(ImageFilter.GaussianBlur(radius=2))
    # 白色区域透明可见，黑色区域不可见
    img.putalpha(alpha)
    return img


def info_calc(
    n1: Optional[float], n2: Optional[float], rank: bool = False, pp: bool = False
):
    if not n1 or not n2:
        return "", 0
    num = n1 - n2
    if num < 0:
        if rank:
            op, value = "↑", num * -1
        elif pp:
            op, value = "↓", num * -1
        else:
            op, value = "-", num * -1
    elif num > 0:
        if rank:
            op, value = "↓", num
        elif pp:
            op, value = "↑", num
        else:
            op, value = "+", num
    else:
        op, value = "", 0
    return [op, value]


def draw_acc(img: Image, acc: float, mode: str):
    acc *= 100
    size = [acc, 100 - acc]
    if mode == "osu":
        insize = [60, 20, 7, 7, 5, 1]
    elif mode == "taiko":
        insize = [60, 20, 5, 5, 4, 1]
    elif mode == "fruits":
        insize = [85, 5, 4, 4, 1, 1]
    else:
        insize = [70, 10, 10, 5, 4, 1]
    insizecolor = ["#ff5858", "#ea7948", "#d99d03", "#72c904", "#0096a2", "#be0089"]
    fig = Figure()
    ax = fig.add_axes((0.1, 0.1, 0.8, 0.8))
    patches = ax.pie(
        size,
        radius=1,
        startangle=90,
        counterclock=False,
        pctdistance=0.9,
        wedgeprops=dict(width=0.20),
        colors=["#66cbfd"],
    )
    ax.pie(
        insize,
        radius=0.8,
        colors=insizecolor,
        startangle=90,
        counterclock=False,
        pctdistance=0.9,
        wedgeprops=dict(width=0.05),
    )
    patches[0][1].set_alpha(0)
    acc_img = BytesIO()
    fig.savefig(acc_img, transparent=True)
    ax.cla()
    ax.clear()
    fig.clf()
    fig.clear()
    score_acc_img = Image.open(acc_img).convert("RGBA").resize((576, 432))
    img.alpha_composite(score_acc_img, (25, 83))
    return img


async def crop_bg(size: str, path: Union[str, Path]):
    try:
        bg = Image.open(path).convert("RGBA")
    except UnidentifiedImageError:
        os.remove(path)
        data = await get_seasonal_bg()
        pic = SeasonalBackgrounds(**data)
        url = random.choice(pic.backgrounds).url
        res = await safe_async_get(url)
        bg = Image.open(BytesIO(res.content)).convert("RGBA")
    except FileNotFoundError:
        data = await get_seasonal_bg()
        pic = SeasonalBackgrounds(**data)
        url = random.choice(pic.backgrounds).url
        res = await safe_async_get(url)
        bg = Image.open(BytesIO(res.content)).convert("RGBA")
    bg_w, bg_h = bg.size[0], bg.size[1]
    if size == "BG":
        fix_w = 1500
        fix_h = 720
    elif size == "H":
        fix_w = 540
        fix_h = 180
    elif size == "HI":
        fix_w = 1000
        fix_h = 400
    elif size == "MP":
        fix_w = 1200
        fix_h = 300
    elif size == "MB":
        fix_w = 1200
        fix_h = 600
    else:
        raise
    # 固定比例
    fix_scale = fix_h / fix_w
    # 图片比例
    bg_scale = bg_h / bg_w
    # 当图片比例大于固定比例
    if bg_scale > fix_scale:
        # 长比例
        scale_width = fix_w / bg_w
        # 等比例缩放
        width = int(scale_width * bg_w)
        height = int(scale_width * bg_h)
        sf = bg.resize((width, height))
        # 计算上下裁切
        crop_height = (height - fix_h) // 2
        x1, y1, x2, y2 = 0, crop_height, width, height - crop_height
        # 裁切保存
        crop_img = sf.crop((x1, y1, x2, y2))
        return crop_img
    # 当图片比例小于固定比例
    elif bg_scale < fix_scale:
        # 宽比例
        scale_height = fix_h / bg_h
        # 等比例缩放
        width = int(scale_height * bg_w)
        height = int(scale_height * bg_h)
        sf = bg.resize((width, height))
        # 计算左右裁切
        crop_width = (width - fix_w) // 2
        x1, y1, x2, y2 = crop_width, 0, width - crop_width, height
        # 裁切保存
        crop_img = sf.crop((x1, y1, x2, y2))
        return crop_img
    else:
        sf = bg.resize((fix_w, fix_h))
        return sf


def stars_diff(stars: float):
    if stars < 0.1:
        r, g, b = 170, 170, 170
    elif stars >= 9:
        r, g, b = 0, 0, 0
    else:
        # 颜色取色参考 https://github.com/ppy/osu-web/blob/97997d9c7b7f9c49f9b3cdd776c71afb9872c34b/resources/js/utils/beatmap-helper.ts#L20
        r, g, b, a = ColorArr[int(stars * 100)]
    # 打开底图
    xx, yy = Stars.size
    # 填充背景
    img = Image.new("RGBA", Stars.size, (r, g, b))
    img.paste(Stars, (0, 0, xx, yy), Stars)
    # 把白色变透明
    arr = np.array(img)
    # 创建mask，将白色替换为True，其他颜色替换为False
    mask = (arr[:, :, 0] == 255) & (arr[:, :, 1] == 255) & (arr[:, :, 2] == 255)
    # 将mask中为True的像素点的alpha通道设置为0
    arr[:, :, 3][mask] = 0
    # 将numpy数组转换回PIL图片
    img = Image.fromarray(arr)
    return img


def get_modeimage(mode: int) -> Path:
    if mode == 0:
        img = "pfm_std.png"
    elif mode == 1:
        img = "pfm_taiko.png"
    elif mode == 2:
        img = "pfm_ctb.png"
    else:
        img = "pfm_mania.png"
    return osufile / img


def calc_songlen(length: int) -> str:
    map_len = list(divmod(int(length), 60))
    map_len[1] = map_len[1] if map_len[1] >= 10 else f"0{map_len[1]}"
    music_len = f"{map_len[0]}:{map_len[1]}"
    return music_len


async def open_user_icon(info: User) -> Image:
    path = user_cache_path / str(info.id)
    png_user_icon = user_cache_path / str(info.id) / "icon.png"
    gif_user_icon = user_cache_path / str(info.id) / "icon.gif"
    # 判断文件是否存在，并读取图片
    if png_user_icon.exists():
        image = Image.open(png_user_icon)
    elif gif_user_icon.exists():
        image = Image.open(gif_user_icon)
    else:
        user_icon = await get_projectimg(info.avatar_url)
        with open(path / f"icon.{info.avatar_url[-3:]}", "wb") as f:
            f.write(user_icon.getvalue())
        image = Image.open(user_icon)
    return image


def is_close(n1, n2) -> bool:
    if abs(n1 - n2) < 0.01:
        return True
    return False



from datetime import datetime, timedelta
from io import BytesIO
from typing import Union
from PIL import ImageEnhance, ImageDraw

from ..api import osu_api, get_map_bg
from ..beatmap_stats_moder import with_mods
from ..schema import Beatmap
from ..mods import calc_mods
from ..file import re_map, get_projectimg, download_osu, map_path
from ..utils import FGM
from ..pp import get_ss_pp
from .utils import calc_songlen, draw_fillet, stars_diff, crop_bg, image2bytesio, is_close
from .static import *


async def draw_map_info(mapid: int, mods: list) -> Union[str, BytesIO]:
    info = await osu_api("map", map_id=mapid)
    if not info:
        return "未查询到该地图信息"
    if isinstance(info, str):
        return info
    mapinfo = Beatmap(**info)
    original_mapinfo = mapinfo.copy()
    mapinfo = with_mods(mapinfo, None, mods)
    diffinfo = (
        calc_songlen(mapinfo.total_length),
        f"{mapinfo.bpm:.1f}",
        mapinfo.count_circles,
        mapinfo.count_sliders,
    )
    # 获取地图
    path = map_path / str(mapinfo.beatmapset_id)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    osu = path / f"{mapid}.osu"
    if not osu.exists():
        await download_osu(mapinfo.beatmapset_id, mapid)
    ss_pp_info = get_ss_pp(str(osu.absolute()), calc_mods(mods))
    original_ss_pp_info = get_ss_pp(str(osu.absolute()), 0)
    # 计算时间
    if mapinfo.beatmapset.ranked_date:
        old_time = datetime.strptime(
            mapinfo.beatmapset.ranked_date.replace("Z", ""), "%Y-%m-%dT%H:%M:%S"
        )
        new_time = (old_time + timedelta(hours=8)).strftime("%Y-%m-%d %H:%M:%S")
    else:
        new_time = "谱面状态非上架"
    # BG做地图
    im = Image.new("RGBA", (1200, 600))
    draw = ImageDraw.Draw(im)
    cover = re_map(osu)
    cover_path = path / cover
    if not cover_path.exists():
        if bg := await get_map_bg(mapid, mapinfo.beatmapset_id, cover):
            with open(cover_path, "wb") as f:
                f.write(bg.getvalue())
    cover_crop = await crop_bg("MB", cover_path)
    cover_img = ImageEnhance.Brightness(cover_crop).enhance(2 / 4.0)
    im.alpha_composite(cover_img)
    # 获取地图info
    if FGM[mapinfo.mode] == 3:
        im.alpha_composite(MapBg1)
    else:
        im.alpha_composite(MapBg)
    # 模式
    draw.text((50, 65), IconLs[mapinfo.mode], font=extra_30, anchor="lt")
    # 难度星星
    stars_bg = stars_diff(ss_pp_info.difficulty.stars)
    stars_img = stars_bg.resize((80, 30))
    im.alpha_composite(stars_img, (90, 65))
    if ss_pp_info.difficulty.stars < 6.5:
        color = (0, 0, 0, 255)
    else:
        color = (255, 217, 102, 255)
    # 星级
    draw.text(
        (100, 78),
        f"★{ss_pp_info.difficulty.stars:.2f}",
        font=Torus_SemiBold_20,
        anchor="lm",
        fill=color,
    )
    # cs, ar, od, hp
    mapdiff = [mapinfo.cs, mapinfo.drain, mapinfo.accuracy, mapinfo.ar]
    original_mapdiff = [
        original_mapinfo.cs,
        original_mapinfo.drain,
        original_mapinfo.accuracy,
        original_mapinfo.ar,
    ]

    for num, (original, new) in enumerate(zip(original_mapdiff, mapdiff)):
        if is_close(new, original):
            color = (255, 255, 255, 255)
            orig_difflen = int(250 * max(0, original) / 10) if original <= 10 else 250
            orig_diff_len = Image.new("RGBA", (orig_difflen, 8), color)
            im.alpha_composite(orig_diff_len, (890, 426 + 35 * num))
        elif new > original:
            color = (198, 92, 102, 255)
            orig_color = (246, 136, 144, 255)
            new_difflen = int(250 * max(0, new) / 10) if new <= 10 else 250
            new_diff_len = Image.new("RGBA", (new_difflen, 8), color)
            im.alpha_composite(new_diff_len, (890, 426 + 35 * num))
            orig_difflen = int(250 * max(0, original) / 10) if original <= 10 else 250
            orig_diff_len = Image.new("RGBA", (orig_difflen, 8), orig_color)
            im.alpha_composite(orig_diff_len, (890, 426 + 35 * num))
        elif new < original:
            color = (161, 212, 238, 255)
            orig_color = (255, 255, 255, 255)
            orig_difflen = int(250 * max(0, original) / 10) if original <= 10 else 250
            orig_diff_len = Image.new("RGBA", (orig_difflen, 8), orig_color)
            im.alpha_composite(orig_diff_len, (890, 426 + 35 * num))
            new_difflen = int(250 * max(0, new) / 10) if new <= 10 else 250
            new_diff_len = Image.new("RGBA", (new_difflen, 8), color)
            im.alpha_composite(new_diff_len, (890, 426 + 35 * num))
        else:
            raise Exception('没有这种情况')
        draw.text(
            (1170, 428 + 35 * num),
            str(float("%.2f" % new)).rstrip("0").rstrip("."),
            font=Torus_SemiBold_20,
            anchor="mm",
        )
    # stardiff
    stars = ss_pp_info.difficulty.stars
    original_stars = original_ss_pp_info.difficulty.stars
    if stars > original_stars:
        color = (198, 92, 102, 255)
        orig_color = (246, 111, 34, 255)
        new_difflen = int(250 * max(0.0, stars) / 10) if stars <= 10 else 250
        new_diff_len = Image.new("RGBA", (new_difflen, 8), color)
        im.alpha_composite(new_diff_len, (890, 566))
        orig_difflen = (
            int(250 * max(0.0, original_stars) / 10) if original_stars <= 10 else 250
        )
        orig_diff_len = Image.new("RGBA", (orig_difflen, 8), orig_color)
        im.alpha_composite(orig_diff_len, (890, 566))
    elif stars < original_stars:
        color = (161, 187, 127, 255)
        orig_color = (255, 204, 34, 255)
        orig_difflen = (
            int(250 * max(0.0, original_stars) / 10) if original_stars <= 10 else 250
        )
        orig_diff_len = Image.new("RGBA", (orig_difflen, 8), orig_color)
        im.alpha_composite(orig_diff_len, (890, 566))
        new_difflen = int(250 * max(0.0, stars) / 10) if stars <= 10 else 250
        new_diff_len = Image.new("RGBA", (new_difflen, 8), color)
        im.alpha_composite(new_diff_len, (890, 566))
    else:
        color = (255, 204, 34, 255)
        difflen = int(250 * stars / 10) if stars <= 10 else 250
        diff_len = Image.new("RGBA", (difflen, 8), color)
        im.alpha_composite(diff_len, (890, 566))
    draw.text((1170, 566), f"{stars:.2f}", font=Torus_SemiBold_20, anchor="mm")
    # 绘制mods
    if mods:
        for mods_num, s_mods in enumerate(mods):
            mods_bg = osufile / "mods" / f"{s_mods}.png"
            mods_img = Image.open(mods_bg).convert("RGBA")
            im.alpha_composite(mods_img, (700 + 50 * mods_num, 295))
    # mapper
    icon_url = f"https://a.ppy.sh/{mapinfo.user_id}"
    user_icon = await get_projectimg(icon_url)
    if not user_icon.exists():
        return '用户头像下载出错，请重试！'
    icon = Image.open(user_icon).convert("RGBA").resize((100, 100))
    icon_img = draw_fillet(icon, 10)
    im.alpha_composite(icon_img, (50, 400))
    # mapid
    draw.text(
        (1190, 80),
        f"Setid: {mapinfo.beatmapset_id} | Mapid: {mapid}",
        font=Torus_Regular_20,
        anchor="rm",
    )
    # 难度名
    version = mapinfo.version
    max_version_length = 60
    if len(version) > max_version_length:
        version = version[: max_version_length - 3] + "..."
    text = f"{version}"
    draw.text((180, 78), text, font=Torus_SemiBold_20, anchor="lm")
    # 曲名+曲师
    text = f"{mapinfo.beatmapset.title} | by {mapinfo.beatmapset.artist_unicode}"
    max_length = 80
    if len(text) > max_length:
        text = text[: max_length - 3] + "..."

    draw.text((50, 37), text, font=Torus_SemiBold_25, anchor="lm")
    # 来源
    # draw.text((50, 260), f"Source:{mapinfo.beatmapset.source}", font=Torus_SemiBold_25, anchor="rt")
    # mapper
    draw.text((160, 400), "谱师:", font=Torus_SemiBold_20, anchor="lt")
    draw.text(
        (160, 425), mapinfo.beatmapset.creator, font=Torus_SemiBold_20, anchor="lt"
    )
    # ranked时间
    draw.text((160, 460), "上架时间:", font=Torus_SemiBold_20, anchor="lt")
    draw.text((160, 485), new_time, font=Torus_SemiBold_20, anchor="lt")
    # 状态
    draw.text(
        (1100, 304), mapinfo.status.capitalize(), font=Torus_SemiBold_20, anchor="mm"
    )
    # 时长 - 滑条
    for num, i in enumerate(diffinfo):
        draw.text(
            (770 + 120 * num, 365),
            f"{i}",
            font=Torus_Regular_20,
            anchor="lm",
            fill=(255, 204, 34, 255),
        )
    # maxcb
    draw.text(
        (50, 570), f"最大连击: {mapinfo.max_combo}", font=Torus_SemiBold_20, anchor="lm"
    )
    # pp
    draw.text(
        (320, 570),
        f"SS PP: {int(round(ss_pp_info.pp, 0))}",
        font=Torus_SemiBold_20,
        anchor="lm",
    )
    # 输出
    base = image2bytesio(im)
    im.close()
    return base



from .guess import (
    guess_audio,
    guess_pic,
    word_matcher,
    pic_word_matcher,
    hint,
    pic_hint,
)
from .medal import medal
# from .rank import group_pp_rank

# from .bp_analyze import bp_analyze
from .pr import pr, recent
from .osu_help import osu_help
from .url_match import url_match
from .recommend import recommend
from .update import update_info, update_pic, clear_background
from .preview import generate_preview
from .getbg import getbg
from .bind import bind, unbind
from .bp import bp, tbp
from .info import info
from .map import osu_map, bmap
from .mu import mu
from .score import score
from .update_mode import update_mode
from .history import history
from .map_convert import convert, change, generate_full_ln



from pathlib import Path
from PIL import Image, ImageFont
import matplotlib.colors as mcolors
import numpy as np
import matplotlib as mpl

osufile = Path(__file__).parent.parent / "osufile"
Torus_Regular_20 = ImageFont.truetype(str(osufile / "fonts" / "Torus Regular.otf"), 20)
Torus_Regular_25 = ImageFont.truetype(str(osufile / "fonts" / "Torus Regular.otf"), 25)
Torus_Regular_30 = ImageFont.truetype(str(osufile / "fonts" / "Torus Regular.otf"), 30)
Torus_Regular_35 = ImageFont.truetype(str(osufile / "fonts" / "Torus Regular.otf"), 35)
Torus_Regular_40 = ImageFont.truetype(str(osufile / "fonts" / "Torus Regular.otf"), 40)
Torus_Regular_45 = ImageFont.truetype(str(osufile / "fonts" / "Torus Regular.otf"), 45)
Torus_Regular_50 = ImageFont.truetype(str(osufile / "fonts" / "Torus Regular.otf"), 50)
Torus_Regular_75 = ImageFont.truetype(str(osufile / "fonts" / "Torus Regular.otf"), 75)
Torus_SemiBold_20 = ImageFont.truetype(
    str(osufile / "fonts" / "Torus SemiBold.otf"), 20
)
Torus_SemiBold_25 = ImageFont.truetype(
    str(osufile / "fonts" / "Torus SemiBold.otf"), 25
)
Torus_SemiBold_30 = ImageFont.truetype(
    str(osufile / "fonts" / "Torus SemiBold.otf"), 30
)
Torus_SemiBold_40 = ImageFont.truetype(
    str(osufile / "fonts" / "Torus SemiBold.otf"), 40
)
Torus_SemiBold_50 = ImageFont.truetype(
    str(osufile / "fonts" / "Torus SemiBold.otf"), 50
)
Venera_75 = ImageFont.truetype(str(osufile / "fonts" / "Venera.otf"), 75)
extra_30 = ImageFont.truetype(str(osufile / "fonts" / "Extra.otf"), 30)

InfoImg = Image.open(osufile / "info.png").convert("RGBA")
SupporterBg = Image.open(osufile / "work" / "suppoter.png").convert("RGBA")
ExpLeftBg = Image.open(osufile / "work" / "left.png").convert("RGBA")
ExpCenterBg = Image.open(osufile / "work" / "center.png").convert("RGBA")
ExpRightBg = Image.open(osufile / "work" / "right.png").convert("RGBA")
BgImg = Image.open(osufile / "Best Performance.png").convert("RGBA")
BgImg1 = Image.open(osufile / "History Score.jpg").convert("RGBA")
MapBg = Image.open(osufile / "beatmapinfo.png").convert("RGBA")
MapBg1 = Image.open(osufile / "maniabeatmapinfo.png").convert("RGBA")
BarImg = Image.open(osufile / "work" / "bmap.png").convert("RGBA")
Stars = Image.open(osufile / "work" / f"stars.png").convert("RGBA")

# 颜色取色参考 https://github.com/ppy/osu-web/blob/97997d9c7b7f9c49f9b3cdd776c71afb9872c34b/resources/js/utils/beatmap-helper.ts#L20

__input_values = np.array([0.1, 1.25, 2, 2.5, 3.3, 4.2, 4.9, 5.8, 6.7, 7.7, 9])
__normalized_values = (__input_values - np.min(__input_values)) / (np.max(__input_values) - np.min(__input_values))

# 定义对应的颜色
__colors = ['#4290FB', '#4FC0FF', '#4FFFD5', '#7CFF4F', '#F6F05C', '#FF8068', '#FF4E6F', '#C645B8', '#6563DE', '#18158E', '#000000']

# 创建颜色映射对象
__cmap = mcolors.LinearSegmentedColormap.from_list('difficultyColourSpectrum', list(zip(__normalized_values, __colors)), N=16384)
__norm = mpl.colors.Normalize(vmin=0, vmax=9)
ColorArr = mpl.cm.ScalarMappable(norm=__norm, cmap=__cmap).to_rgba(np.linspace(0, 9, 900), bytes=True)
IconLs = {"osu": '\uE800', "taiko": '\uE803', "fruits": '\uE801', "mania": '\uE802'}



from collections import defaultdict
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import pandas as pd
from nonebot import on_command
from nonebot.typing import T_State


from .utils import split_msg
from ..utils import NGM
from ..schema import Score
from ..api import safe_async_get, osu_api

bp_analyze = on_command("bpa", aliases={"bp分析"}, priority=11, block=True)
plt.rcParams["font.sans-serif"] = ["SimHei"]
plt.rcParams["axes.unicode_minus"] = False
rank_color = {
    "X": "#ffe193",
    "XH": "#ffe193",
    "S": "#ffe193",
    "SH": "#ffe193",
    "A": "#88da20",
    "B": "#e9b941",
    "C": "#fa8a59",
    "D": "#f55757",
}


@bp_analyze.handle(parameterless=[split_msg()])
async def _(state: T_State):
    user = state["para"] if state["para"] else state["user"]
    bp_info = await osu_api("bp", user, NGM[state["mode"]], is_name=state["is_name"])
    if not bp_info:
        return f'未查询到在 {NGM[state["mode"]]} 的游玩记录'
    if isinstance(bp_info, str):
        return bp_info
    score_ls = [Score(**i) for i in bp_info]
    song_length = pd.Series([i.beatmap.total_length for i in score_ls])
    colors = [rank_color[i.rank] for i in score_ls]
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    ax2 = ax1.twinx()
    # 绘制柱状图
    ax1.bar(
        range(len(song_length)), song_length, color=colors, width=0.8, align="center"
    )
    # 添加柱状图边框
    for num, (i, j) in enumerate(zip(range(len(song_length)), song_length)):
        if score_ls[num].rank in ("X", "XH"):
            rect = plt.Rectangle(
                (i - 0.4, 0), 0.8, j, edgecolor="yellow", facecolor="none", linewidth=2
            )
            ax1.add_patch(rect)
    # 设置 y 轴刻度为 xx:xx 的格式
    ax2.plot([i.pp for i in score_ls])
    ax1.yaxis.set_major_locator(ticker.MultipleLocator(base=60))  # 设置主刻度为每分钟
    ax1.yaxis.set_major_formatter(
        ticker.FuncFormatter(lambda x, pos: f"{x // 60:.0f}分钟")
    )
    plt.show()
    # 统计mods pp
    mods_pp = defaultdict(int)
    for num, i in enumerate(score_ls):
        if not i.mods:
            mods_pp["NM"] += i.pp * 0.95**num
        for j in i.mods:
            mods_pp[j] += i.pp * 0.95**num
    plt.pie(mods_pp.values(), labels=mods_pp.keys())
    legend_labels = [
        f"{label}: {size:.0f}pp"
        for label, size in zip(mods_pp.keys(), mods_pp.values())
    ]
    plt.legend(legend_labels, loc="upper right")
    plt.show()



import asyncio

from httpx import AsyncClient
from nonebot import get_plugin_config
from ..config import Config

plugin_config = get_plugin_config(Config)
proxy = plugin_config.osu_proxy


async def fetch_url(client: AsyncClient, url):
    try:
        response = await client.get(url)
        return response
    except Exception:
        return None


async def get_first_response(urls: list[str]):
    async with AsyncClient(proxies=proxy, follow_redirects=True) as client:
        tasks = [asyncio.create_task(fetch_url(client, url)) for url in urls]
        while tasks:
            done, _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
            for task in done:
                response = task.result()
                if response is not None and response.status_code == 200:
                    return response.content
            tasks = [task for task in tasks if not task.done()]
        return None



from io import BytesIO

import matplotlib.pyplot as plt
import datetime
from nonebot import on_command
from nonebot_plugin_alconna import UniMessage
from nonebot.typing import T_State
from .utils import split_msg
from ..utils import NGM
from ..database import InfoData, UserData

history = on_command("history", priority=11, block=True)


@history.handle(parameterless=[split_msg()])
async def _info(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).finish(reply_to=True)
    data = InfoData.filter(osu_id=state["user"], osu_mode=state["mode"])
    user = await UserData.filter(osu_id=state["user"]).first()
    if state["day"] > 0:
        data = data.filter(
            date__gte=datetime.date.today() - datetime.timedelta(days=state["day"])
        )
    data = await data.order_by("date").all()
    pp_ls = [i.pp for i in data]
    date_ls = [i.date for i in data]
    rank_ls = [i.g_rank for i in data]
    byt = draw_plot(
        pp_ls, date_ls, rank_ls, f'{user.osu_name} {NGM[state["mode"]]} pp/rank history'
    )
    await UniMessage.image(raw=byt).finish(reply_to=True)


def draw_plot(pp_ls, date_ls, rank_ls, title):
    fig = plt.figure(num="draw_double_line", figsize=(20, 6))
    ax1 = fig.add_subplot(111)
    ax1.plot(date_ls, pp_ls, "black", label="pp")
    ax1.legend(loc="upper left")
    ax2 = ax1.twinx()
    ax2.plot(date_ls, rank_ls, "r", label="rank")
    ax2.legend(loc="upper right")
    ax2.invert_yaxis()
    plt.grid()
    plt.xlabel("date", fontsize=15)
    plt.title(title)
    byt = BytesIO()
    plt.savefig(byt, format="png")
    plt.close()
    return byt



from nonebot import on_command
from nonebot.internal.adapter import Message
from nonebot.params import CommandArg
from nonebot_plugin_alconna import UniMessage
from ..file import download_map

osudl = on_command("osudl", priority=11, block=True)


@osudl.handle()
async def _osudl(setid: Message = CommandArg()):
    setid = setid.extract_plain_text().strip()
    if not setid or not setid.isdigit():
        await UniMessage.text("请输入正确的地图ID").send(reply_to=True)
    osz_path = await download_map(int(setid))
    file_path = osz_path.absolute()
    try:
        await UniMessage.file(path=file_path).send()
    except Exception as e:
        await UniMessage.text("上传文件失败，可能是群空间满或没有权限导致的").send(reply_to=True)
    finally:
        try:
            osz_path.unlink()
        except PermissionError:
            ...



from nonebot import on_command
from nonebot_plugin_alconna import UniMessage
from nonebot.typing import T_State
from .utils import split_msg
from ..draw import draw_score, draw_bp
from ..utils import NGM

bp = on_command("bp", priority=11, block=True)

pfm = on_command("pfm", priority=11, block=True)
tbp = on_command("tbp", priority=11, block=True, aliases={"todaybp"})


@bp.handle(parameterless=[split_msg()])
async def _bp(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).finish(reply_to=True)
    if "-" in state["para"]:
        await _pfm(state)
        return
    best = state["para"]
    if not best:
        best = "1"
    if not best.isdigit():
        await UniMessage.text("只能接受纯数字的bp参数").finish(reply_to=True)
    best = int(best)
    if best <= 0 or best > 100:
        await UniMessage.text("只允许查询bp 1-100 的成绩").finish(reply_to=True)
    data = await draw_score(
        "bp",
        state["user"],
        NGM[state["mode"]],
        best=best,
        mods=state["mods"],
        is_name=state["is_name"],
    )
    if isinstance(data, str):
        await UniMessage.text(data).finish(reply_to=True)
    await UniMessage.image(raw=data).finish(reply_to=True)


@pfm.handle(parameterless=[split_msg()])
async def _pfm(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).finish(reply_to=True)
    ls = state["para"].split("-")
    low, high = ls[0], ls[1]
    if not low.isdigit() or not high.isdigit():
        await UniMessage.text('参数应为 "数字-数字"的形式!').finish(reply_to=True)
    low, high = int(low), int(high)
    if not 0 < low < high <= 100:
        await UniMessage.text("仅支持查询bp1-100").finish(reply_to=True)
    data = await draw_bp(
        "bp",
        state["user"],
        NGM[state["mode"]],
        state["mods"],
        low,
        high,
        is_name=state["is_name"],
    )
    if isinstance(data, str):
        await UniMessage.text(data).finish(reply_to=True)
    await UniMessage.image(raw=data).finish(reply_to=True)


@tbp.handle(parameterless=[split_msg()])
async def _tbp(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).finish(reply_to=True)
    data = await draw_bp(
        "tbp",
        state["user"],
        NGM[state["mode"]],
        [],
        day=state["day"],
        is_name=state["is_name"],
    )
    if isinstance(data, str):
        await UniMessage.text(data).finish(reply_to=True)
    await UniMessage.image(raw=data).finish(reply_to=True)



from nonebot import on_command
from nonebot_plugin_alconna import UniMessage
from nonebot.params import T_State

from ..api import get_user_info
from .utils import split_msg
from ..draw import draw_score
from ..schema import Score
from ..api import osu_api
from ..draw.bp import draw_pfm
from ..utils import NGM

recent = on_command("recent", priority=11, block=True, aliases={"re", "RE", "Re", "rE"})
pr = on_command("pr", priority=11, block=True, aliases={"PR", "Pr", "pR"})


@recent.handle(parameterless=[split_msg()])
async def _recent(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).send(reply_to=True)
        return
    mode = NGM[state["mode"]]
    if "-" in state["para"]:
        ls = state["para"].split("-")
        low, high = ls[0], ls[1]
        data = await osu_api(
            "recent",
            state["user"],
            mode,
            is_name=state["is_name"],
            offset=low,
            limit=high,
        )
        if not data:
            await UniMessage.text(f"未查询到在 {mode} 的游玩记录").send(reply_to=True)
            return
        if isinstance(data, str):
            await UniMessage.text(data).send(reply_to=True)
        if not state["is_name"]:
            info = await get_user_info(
                f"https://osu.ppy.sh/api/v2/users/{state['user']}?key=id"
            )
            if isinstance(info, str):
                return info
            else:
                state["user"] = info["username"]
        score_ls = [Score(**score_json) for score_json in data]
        pic = await draw_pfm("relist", state["user"], score_ls, score_ls, mode)
        await UniMessage.image(raw=pic).send(reply_to=True)
        return
    if state["day"] == 0:
        state["day"] = 1
    data = await draw_score(
        "recent", state["user"], mode, [], state["day"] - 1, is_name=state["is_name"]
    )
    if isinstance(data, str):
        await UniMessage.text(data).send(reply_to=True)
        return
    await UniMessage.image(raw=data).send(reply_to=True)


@pr.handle(parameterless=[split_msg()])
async def _pr(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).send(reply_to=True)
        return
    mode = state["mode"]
    if "-" in state["para"]:
        ls = state["para"].split("-")
        low, high = ls[0], ls[1]
        data = await osu_api(
            "pr",
            state["user"],
            NGM[mode],
            is_name=state["is_name"],
            offset=low,
            limit=high,
        )
        if not data:
            await UniMessage.text(f"未查询到在 {NGM[mode]} 的游玩记录").finish(reply_to=True)
        if isinstance(data, str):
            await UniMessage.text(data).finish(reply_to=True)
        if not state["is_name"]:
            info = await get_user_info(
                f"https://osu.ppy.sh/api/v2/users/{state['user']}?key=id"
            )
            if isinstance(info, str):
                return info
            else:
                state["user"] = info["username"]
        score_ls = [Score(**score_json) for score_json in data]
        pic = await draw_pfm("prlist", state["user"], score_ls, score_ls, NGM[mode])
        await UniMessage.image(raw=pic).finish(reply_to=True)
    if state["day"] == 0:
        state["day"] = 1
    data = await draw_score(
        "pr", state["user"], NGM[mode], [], state["day"] - 1, is_name=state["is_name"]
    )
    if isinstance(data, str):
        await UniMessage.text(data).finish(reply_to=True)
    await UniMessage.image(raw=data).finish(reply_to=True)



from nonebot import on_command
from nonebot_plugin_alconna import UniMessage
from nonebot.typing import T_State
from .utils import split_msg
from ..draw import draw_info
from ..utils import NGM

info = on_command("info", priority=11, block=True, aliases={"Info", "INFO"})


@info.handle(parameterless=[split_msg()])
async def _info(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).finish(reply_to=True)
    data = await draw_info(
        state["user"], NGM[state["mode"]], state["day"], state["is_name"]
    )
    if isinstance(data, str):
        await UniMessage.text(data).finish(reply_to=True)
    await UniMessage.image(raw=data).finish(reply_to=True)



from nonebot import on_command
from nonebot_plugin_alconna import UniMessage
from nonebot.typing import T_State

from .utils import split_msg
from ..draw import draw_map_info, draw_bmap_info

osu_map = on_command("map", priority=11, block=True)
bmap = on_command("bmap", priority=11, block=True)


@osu_map.handle(parameterless=[split_msg()])
async def _map(state: T_State):
    map_id = state["para"]
    mods = state["mods"]
    if not map_id:
        await UniMessage.text("请输入地图ID").finish(reply_to=True)
    elif not map_id.isdigit():
        await UniMessage.text("请输入正确的地图ID").finish(reply_to=True)
    m = await draw_map_info(map_id, mods)
    if isinstance(m, str):
        await UniMessage.text(m).finish(reply_to=True)
    await UniMessage.image(raw=m).finish(reply_to=True)


@bmap.handle(parameterless=[split_msg()])
async def _bmap(state: T_State):
    set_id = state["para"]
    if not set_id:
        await UniMessage.text("请输入setID").finish(reply_to=True)
    if not set_id.isdigit():
        await UniMessage.text("请输入正确的setID").finish(reply_to=True)
    m = await draw_bmap_info(set_id)
    if isinstance(m, str):
        await UniMessage.text(m).finish(reply_to=True)
    await UniMessage.image(raw=m).finish(reply_to=True)



from nonebot.internal.adapter import Event, Message
from nonebot.internal.params import Depends
from nonebot.params import T_State, CommandArg
from nonebot_plugin_alconna import UniMsg, At

from ..utils import mods2list
from ..database.models import UserData


def split_msg():
    async def dependency(
        event: Event, state: T_State, msg: UniMsg, arg: Message = CommandArg()
    ):
        qq = event.get_user_id()
        if msg.has(At):
            qq = msg.get(At)[0].target
        user_data = await UserData.get_or_none(user_id=int(qq))
        state["user"] = user_data.osu_id if user_data else 0
        state["mode"] = str(user_data.osu_mode) if user_data else "0"
        state["mods"] = []
        state["day"] = 0
        state["is_name"] = False
        symbol_ls = [":", "+", "：", "#", "＃"]
        symbol_dic = {":": "mode", "+": "mods", "：": "mode", "#": "day", "＃": "day"}
        double_command = ("bp", "score")
        dic = {}
        arg = arg.extract_plain_text().strip()
        if max([arg.find(i) for i in symbol_ls]) >= 0:
            for i in symbol_ls:
                dic[i] = arg.find(i)
            sorted_dict = sorted(dic.items(), key=lambda x: x[1])
            for i in range(len(sorted_dict) - 1):
                if sorted_dict[i][1] >= 0:
                    state[symbol_dic[sorted_dict[i][0]]] = arg[
                        sorted_dict[i][1] + 1 : sorted_dict[i + 1][1]
                    ].strip()
            if sorted_dict[-1][1] >= 0:
                state[symbol_dic[sorted_dict[-1][0]]] = arg[
                    sorted_dict[-1][1] + 1 :
                ].strip()
            if isinstance(state["mods"], str):
                state["mods"] = mods2list(state["mods"].strip())
            index = min([arg.find(i) for i in symbol_ls if arg.find(i) >= 0])
            state["para"] = arg[:index].strip()
        else:
            state["para"] = arg.strip()
        if " " in state["para"]:
            ls = state["para"].split(" ")
            if state["_prefix"]["command"][0] in double_command:
                state["para"] = ls[-1]
                state["user"] = " ".join(ls[:-1])
            elif is_num_hyphen_num(ls[-1]):
                state["para"] = ls[-1]
                state["user"] = " ".join(ls[:-1])
            elif is_num_hyphen_num(ls[0]):
                state["para"] = ls[0]
                state["user"] = " ".join(ls[1:])
            else:
                state["user"] = state["para"]
            state["is_name"] = True
        elif state["para"]:
            if (
                not is_num_hyphen_num(state["para"])
                and state["_prefix"]["command"][0] not in double_command
            ):
                state["user"] = state["para"]
                state["is_name"] = True
        # 判断参数是否合法
        if not state["mode"].isdigit() or (
            int(state["mode"]) < 0 or int(state["mode"]) > 3
        ):
            state["error"] = "模式应为0-3的数字！"
        if (
            isinstance(state["day"], str)
            and not state["day"].isdigit()
            and (int(state["day"]) < 0)
        ):
            state["error"] = "查询的日期应是一个正数"
        if state["user"] == 0 and not state["user"]:
            state["error"] = "该账号尚未绑定，请输入 /bind 用户名 绑定账号"
        state["day"] = int(state["day"])

    return Depends(dependency)


def is_num_hyphen_num(s: str):
    if s.count("-") != 1:
        return False
    s = s.split("-")
    return s[0].isdigit() and s[1].isdigit()



import re
import json
from pathlib import Path

from nonebot_plugin_alconna import UniMessage
from nonebot import on_command
from nonebot.internal.adapter import Message
from nonebot.params import CommandArg
from ..api import safe_async_get


medal_data_path = Path(__file__).parent.parent / "osufile" / "medals" / "medals.json"
with open(medal_data_path, "r") as file:
    medal_json = json.load(file)


medal = on_command("medal", aliases={"成就"}, priority=11, block=True)


@medal.handle()
async def _(msg: Message = CommandArg()):
    name = msg.extract_plain_text()
    data = await safe_async_get(
        f"https://osekai.net/medals/api/public/get_medal.php?medal={name}"
    )
    medal_data = data.json()
    if "MedalID" not in medal_data:
        await medal.finish("没有找到欸，看看是不是名字打错了")
    words = ""
    if medal_data["Restriction"] != "NULL":
        words += f"限制模式：{medal_data['Restriction']}\n"
    words += "获得方式：\n"
    if medal_data["Name"] in medal_json:
        words += medal_json[medal_data["Name"]]["MedalSolution"]
    else:
        words += (
            medal_data["Solution"]
            if medal_data["Solution"]
            else medal_data["Instructions"]
        )
        table_regex = r"<table[^>]*>(.*?)<\/table>"
        table_match = re.search(table_regex, words, re.DOTALL)
        if table_match:
            table_text = table_match.group(1)

            # 使用正则表达式匹配并提取表格行部分
            row_regex = r"<tr[^>]*>(.*?)<\/tr>"
            rows = re.findall(row_regex, table_text, re.DOTALL)

            result = ""

            # 遍历表格行并提取单元格内容
            for row in rows:
                # 使用正则表达式匹配并提取单元格部分
                cell_regex = r"<t[hd][^>]*>(.*?)<\/t[hd]>"
                cells = re.findall(cell_regex, row, re.DOTALL)
                for cell in cells:
                    # 去除单元格内的HTML标签
                    cell_text = re.sub(r"<[^>]*>", "", cell)
                    result += cell_text + " "
                result += "\n"

            # 将提取的表格文字替换回原文中
            words = re.sub(table_regex, result, words)
    style_regex = r"<style[^>]*>(.*?)<\/style>"
    words = re.sub(style_regex, "", words)
    words = re.sub(r"<[^>]+>", "", words)
    if medal_data["PackID"]:
        words += (
            f'\nhttps://osu.ppy.sh/beatmaps/packs/{medal_data["PackID"].rstrip(",,,")}'
        )
    await (UniMessage.image(url=medal_data["Link"]) + words).send(reply_to=True)
    if medal_data["beatmaps"]:
        msg = UniMessage()
        if len(medal_data["beatmaps"]) > 5:
            medal_data["beatmaps"] = medal_data["beatmaps"][:5]
        for beatmap in medal_data["beatmaps"]:
            msg += (
                f'{beatmap["SongTitle"]} [{beatmap["DifficultyName"]}]\n{beatmap["Difficulty"]}⭐\n'
                + f'https://osu.ppy.sh/b/{beatmap["BeatmapID"]}'
            )
        await msg.send()



from pydantic import BaseModel
from typing import Optional, Union


class Config(BaseModel):
    osu_client: Optional[int] = None
    osu_key: Optional[str] = None
    info_bg: Optional[list[str]] = ["https://api.hiosu.com/erciyuan/pe.php"]
    osu_proxy: Optional[Union[str, dict]] = None



from nonebot import on_command
from nonebot_plugin_alconna import UniMessage
from nonebot.typing import T_State
from .utils import split_msg
from ..draw import get_score_data
from ..utils import NGM


score = on_command("score", priority=11, block=True)


@score.handle(parameterless=[split_msg()])
async def _score(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).finish(reply_to=True)
    data = await get_score_data(
        state["user"],
        NGM[state["mode"]],
        mapid=state["para"],
        mods=state["mods"],
        is_name=state["is_name"],
    )
    if isinstance(data, str):
        await UniMessage.text(data).finish(reply_to=True)
    await UniMessage.image(raw=data).finish(reply_to=True)



from typing import Optional

from .schema import Beatmap, Score

OD0_MS = 80
OD10_MS = 20
AR0_MS = 1800
AR5_MS = 1200
AR10_MS = 450

OD_MS_STEP = (OD0_MS - OD10_MS) / 10
AR_MS_STEP1 = (AR0_MS - AR5_MS) / 5
AR_MS_STEP2 = (AR5_MS - AR10_MS) / 5


def modify_ar(base_ar, speed_mul, multiplier):
    ar = base_ar
    ar *= multiplier

    arms = (
        AR0_MS - AR_MS_STEP1 * ar
        if ar < 5
        else AR5_MS - AR_MS_STEP2 * (ar - 5)
    )

    arms = min(AR0_MS, max(AR10_MS, arms))
    arms /= speed_mul

    ar = (
        (AR0_MS - arms) / AR_MS_STEP1
        if arms > AR5_MS
        else 5 + (AR5_MS - arms) / AR_MS_STEP2
    )
    return ar


def modify_od(base_od, speed_mul, multiplier):
    od = base_od
    od *= multiplier
    odms = OD0_MS - OD_MS_STEP * od
    odms = min(OD0_MS, max(OD10_MS, odms))
    odms /= speed_mul
    od = (OD0_MS - odms) / OD_MS_STEP
    return od


def with_mods(mapinfo: Beatmap, scoreinfo: Optional[Score], mods: list):
    speed_mul = 1
    od_ar_hp_multiplier = 1
    mode = scoreinfo.mode if scoreinfo else mapinfo.mode
    if 'DT' in mods or 'NC' in mods:
        speed_mul = 1.5
        mapinfo.bpm *= 1.5
        mapinfo.total_length /= 1.5
    if 'HT' in mods:
        speed_mul *= 0.75
        mapinfo.bpm *= 0.75
        mapinfo.total_length /= 0.75
    if 'HR' in mods:
        od_ar_hp_multiplier = 1.4
    if 'EZ' in mods:
        od_ar_hp_multiplier *= 0.5
    if mode == "mania":
        speed_mul = 1
    if mode not in ("mania", "taiko"):
        mapinfo.ar = modify_ar(mapinfo.ar, speed_mul, od_ar_hp_multiplier)
    if mode == "fruits":
        speed_mul = 1
    mapinfo.accuracy = modify_od(mapinfo.accuracy, speed_mul, od_ar_hp_multiplier)
    if mode not in ("mania", "taiko"):
        if 'HR' in mods:
            mapinfo.cs *= 1.3
        if 'EZ' in mods:
            mapinfo.cs *= 0.5
        mapinfo.cs = min(10.0, mapinfo.cs)
    mapinfo.drain *= od_ar_hp_multiplier
    mapinfo.drain = min(10.0, mapinfo.drain)
    return mapinfo



import math
from rosu_pp_py import Beatmap, Performance, PerformanceAttributes, GameMode
from .schema import Score
from .mods import calc_mods


def cal_pp(score: Score, path: str) -> PerformanceAttributes:
    beatmap = Beatmap(path=path)
    if score.mode_int == 0:
        mode = GameMode.Osu
    elif score.mode_int == 1:
        mode = GameMode.Taiko
    elif score.mode_int == 2:
        mode = GameMode.Catch
    else:
        mode = GameMode.Mania
    beatmap.convert(mode)
    mods = calc_mods(score.mods)
    if mods & (1 << 9):
        mods -= 1 << 9
        mods += 1 << 6
    if score.mode_int == 2:
        c = Performance(
            accuracy=score.accuracy * 100,
            n_katu=score.statistics.count_katu,
            combo=score.max_combo,
            misses=score.statistics.count_miss,
            n100=score.statistics.count_100,
            n300=score.statistics.count_300,
            mods=mods,
        )
    else:
        c = Performance(
            accuracy=score.accuracy * 100,
            n_katu=score.statistics.count_katu if score.statistics.count_katu else 0,
            n_geki=score.statistics.count_geki if score.statistics.count_geki else 0,
            combo=score.max_combo,
            misses=score.statistics.count_miss if score.statistics.count_miss else 0,
            n50=score.statistics.count_50 if score.statistics.count_50 else 0,
            n100=score.statistics.count_100 if score.statistics.count_100 else 0,
            n300=score.statistics.count_300 if score.statistics.count_300 else 0,
            mods=mods,
        )
    return c.calculate(beatmap)


def get_if_pp_ss_pp(score: Score, path: str) -> tuple:
    beatmap = Beatmap(path=path)
    if score.mode_int == 0:
        mode = GameMode.Osu
    elif score.mode_int == 1:
        mode = GameMode.Taiko
    elif score.mode_int == 2:
        mode = GameMode.Catch
    else:
        mode = GameMode.Mania
    beatmap.convert(mode)
    mods = calc_mods(score.mods)
    if mods & (1 << 9):
        mods -= 1 << 9
        mods += 1 << 6
    c = Performance(
        accuracy=score.accuracy * 100,
        n_katu=score.statistics.count_katu if score.statistics.count_katu else 0,
        n_geki=score.statistics.count_geki if score.statistics.count_geki else 0,
        n50=score.statistics.count_50 if score.statistics.count_50 else 0,
        n100=score.statistics.count_100 if score.statistics.count_100 else 0,
        n300=score.statistics.count_300 + score.statistics.count_miss,
        mods=mods,
    )
    if_pp = c.calculate(beatmap).pp
    c = Performance(accuracy=100, mods=mods)
    ss_pp = c.calculate(beatmap).pp
    if math.isnan(if_pp):
        return "nan", "nan"
    return str(int(round(if_pp, 0))), str(int(round(ss_pp, 0)))


def get_ss_pp(path: str, mods: int) -> PerformanceAttributes:
    beatmap = Beatmap(path=path)
    if mods & (1 << 9):
        mods -= 1 << 9
        mods += 1 << 6
    c = Performance(accuracy=100, mods=mods)
    ss_pp_info = c.calculate(beatmap)
    return ss_pp_info



import shutil
from pathlib import Path
from nonebot.plugin import PluginMetadata, inherit_supported_adapters
from nonebot.log import logger
from nonebot import require

require("nonebot_plugin_apscheduler")
require("nonebot_plugin_alconna")
require("nonebot_plugin_session")
require("nonebot_plugin_tortoise_orm")
from .database.models import UserData
from .info import update_user_info
from .config import Config
from .matcher import *
from nonebot_plugin_apscheduler import scheduler
from nonebot_plugin_tortoise_orm import add_model


add_model("nonebot_plugin_osubot.database.models")
usage = "发送/osuhelp 查看帮助"
__plugin_meta__ = PluginMetadata(
    name="OSUBot",
    description="OSU查分插件",
    usage=usage,
    type="application",
    homepage="https://github.com/yaowan233/nonebot-plugin-osubot",
    config=Config,
    supported_adapters=inherit_supported_adapters("nonebot_plugin_session"),
    extra={
        "unique_name": "osubot",
        "author": "yaowan233 <572473053@qq.com>",
    },
)


@scheduler.scheduled_job("cron", hour="0", misfire_grace_time=60)
async def update_info():
    result = await UserData.all()
    if not result:
        return
    for user in result:
        await update_user_info(user.osu_id)
    logger.info(f"已更新{len(result)}位玩家数据")


@scheduler.scheduled_job("cron", hour="4", day_of_week="0", misfire_grace_time=60)
async def delete_cached_map():
    map_path = Path("data/osu/map")
    shutil.rmtree(map_path)
    map_path.mkdir(parents=True, exist_ok=True)
    user_path = Path("data/osu/user")
    for file_path in user_path.glob("**/*"):
        if file_path.is_file() and file_path.name in ("icon.png", "icon.gif"):
            file_path.unlink()



import random
from io import BytesIO
from httpx import AsyncClient, Response
from typing import Union, Optional
from nonebot.log import logger
from nonebot import get_plugin_config
from expiringdict import ExpiringDict
from .config import Config
from .network import auto_retry
from .network.first_response import get_first_response
from .schema import SayoBeatmap, RecommendData

api = "https://osu.ppy.sh/api/v2"
sayoapi = "https://api.sayobot.cn"
pp_calc_api = "https://api.yuzuai.xyz/osu/ppcalc"
cache = ExpiringDict(max_len=1, max_age_seconds=86400)
plugin_config = get_plugin_config(Config)
proxy = plugin_config.osu_proxy


key = plugin_config.osu_key
client_id = plugin_config.osu_client
bg_url = plugin_config.info_bg


@auto_retry
async def safe_async_get(
    url, headers: Optional[dict] = None, params: Optional[dict] = None
) -> Response:
    async with AsyncClient(proxies=proxy, follow_redirects=True) as client:
        client: AsyncClient
        req = await client.get(url, headers=headers, params=params)
    return req


@auto_retry
async def safe_async_post(url, headers=None, data=None, json=None) -> Response:
    async with AsyncClient(proxies=proxy) as client:
        client: AsyncClient
        req = await client.post(url, headers=headers, data=data, json=json)
    return req


async def renew_token():
    url = "https://osu.ppy.sh/oauth/token"
    if not key or not client_id:
        raise Exception("请设置osu_key和osu_client")
    req = await safe_async_post(
        url,
        json={
            "client_id": f"{client_id}",
            "client_secret": f"{key}",
            "grant_type": "client_credentials",
            "scope": "public",
        }
    )
    if req.status_code == 200:
        osu_token = req.json()
        cache.update({"token": osu_token["access_token"]})
    else:
        logger.error(f"更新OSU token出错 错误{req.status_code}")


async def osu_api(
    project: str,
    uid: int = 0,
    mode: str = "osu",
    map_id: int = 0,
    is_name: bool = False,
    offset: int = 0,
    limit: int = 5,
) -> Union[str, dict]:
    if is_name:
        info = await get_user_info(f"{api}/users/{uid}?key=username")
        if isinstance(info, str):
            return info
        else:
            uid = info["id"]
    if project == "info" or project == "bind" or project == "update":
        url = f"{api}/users/{uid}/{mode}"
    elif project == "recent":
        url = f"{api}/users/{uid}/scores/recent?mode={mode}&include_fails=1&limit={limit}&offset={offset}"
    elif project == "pr":
        url = (
            f"{api}/users/{uid}/scores/recent?mode={mode}&limit={limit}&offset={offset}"
        )
    elif project == "score":
        url = f"{api}/beatmaps/{map_id}/scores/users/{uid}/all?mode={mode}"
    elif project == "bp":
        url = f"{api}/users/{uid}/scores/best?mode={mode}&limit=100"
    elif project == "map":
        url = f"{api}/beatmaps/{map_id}"
    else:
        raise "Project Error"
    return await api_info(project, url)


async def sayo_api(setid: int) -> dict:
    url = f"{sayoapi}/v2/beatmapinfo?0={setid}"
    return await api_info("mapinfo", url)


async def get_user_info(url: str) -> Union[dict, str]:
    token = cache.get("token")
    if not token:
        await renew_token()
        token = cache.get("token")
    header = {"Authorization": f"Bearer {token}"}
    req = await safe_async_get(url, headers=header)
    if not req:
        return "api请求失败，请稍后再试"
    if req.status_code == 401:
        await token.update_token()
        return await get_user_info(url)
    elif req.status_code == 404:
        return "未找到该玩家，请确认玩家ID是否正确，有无多余或缺少的空格"
    elif req.status_code == 200:
        return req.json()
    else:
        return "API请求失败，请联系管理员"


async def api_info(project: str, url: str) -> Union[dict, str]:
    if project == "mapinfo" or project == "PPCalc":
        headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) Chrome/78.0.3904.108"
        }
    else:
        token = cache.get("token")
        if not token:
            await renew_token()
            token = cache.get("token")
        headers = {"Authorization": f"Bearer {token}"}
    req = await safe_async_get(url, headers=headers)
    if not req:
        return "api请求失败，请稍后再试"
    if req.status_code >= 400:
        if project == "info" or project == "bind":
            return "未找到该玩家，请确认玩家ID"
        elif project == "recent":
            return "未找到该玩家，请确认玩家ID"
        elif project == "score":
            return "未找到该地图成绩，请检查是否搞混了mapID与setID或模式"
        elif project == "bp":
            return "未找到该玩家BP"
        elif project == "map":
            return "未找到该地图，请检查是否搞混了mapID与setID"
        else:
            return "API请求失败，请联系管理员或稍后再尝试"
    return req.json()


async def get_beatmap_attribute(map_id, mode):
    url = f"{api}/beatmaps/{map_id}/attributes"
    token = cache.get("token")
    if not token:
        await renew_token()
        token = cache.get("token")
    headers = {"Authorization": f"Bearer {token}"}
    req = await safe_async_post(url, headers, {"ruleset": mode})
    if req.status_code == 404:
        raise "内部错误"
    return req.json()


async def get_random_bg() -> Optional[bytes]:
    res = await safe_async_get(random.choice(bg_url))
    if res.status_code != 200:
        return
    return res.content


async def get_sayo_map_info(sid, t=0) -> SayoBeatmap:
    res = await safe_async_get(f"https://api.sayobot.cn/v2/beatmapinfo?K={sid}&T={t}")
    return SayoBeatmap(**res.json())


async def get_map_bg(mapid, sid, bg_name) -> BytesIO:
    res = await get_first_response([f"https://api.osu.direct/media/background/{mapid}",
                                    f"https://subapi.nerinyan.moe/bg/{mapid}",
                                    f'https://dl.sayobot.cn/beatmaps/files/{sid}/{bg_name}'])
    return BytesIO(res)


async def get_seasonal_bg() -> Optional[dict]:
    url = f"{api}/seasonal-backgrounds"
    token = cache.get("token")
    if not token:
        await renew_token()
        token = cache.get("token")
    headers = {"Authorization": f"Bearer {token}"}
    req = await safe_async_get(url, headers=headers)
    if req.status_code != 200:
        return
    return req.json()


async def get_recommend(uid, mode, key_count):
    headers = {"uid": str(uid)}
    params = {
        "newRecordPercent": "0.2,1",
        "passPercent": "0.2,1",
        "difficulty": "0,15",
        "keyCount": key_count,
        "gameMode": mode,
        "rule": "4",
        "current": "1",
        "pageSize": "20",
        "from": "nonebot_plugin_osubot",
        "hidePlayed": "1",
    }
    res = await safe_async_get(
        "https://alphaosu.keytoix.vip/api/v1/self/maps/recommend",
        headers=headers,
        params=params,
    )
    return RecommendData(**res.json())


async def update_recommend(uid):
    headers = {"uid": str(uid)}
    await safe_async_post(
        "https://alphaosu.keytoix.vip/api/v1/self/users/synchronize", headers=headers
    )



from .score import Score, BeatmapUserScore
from .user import User, Badge
from .beatmap import (
    Beatmap,
    Beatmapset,
    BeatmapDifficultyAttributes,
    SeasonalBackgrounds,
)
from .sayo_beatmap import SayoBeatmap
from .alphaosu import RecommendData



# import asyncio
# import datetime
# from io import BytesIO
#
# from nonebot_plugin_alconna import UniMessage
# from nonebot import on_command
# from nonebot.params import T_State
# from PIL import Image, ImageDraw
# from .utils import split_msg
# from ..database.models import UserData, InfoData
# from ..draw.static import Torus_Regular_25
# from ..draw.utils import draw_fillet
# from ..file import get_projectimg
# from ..utils import NGM
#
#
# group_pp_rank = on_command("群内排名", priority=11, block=True)
#
#
# @group_pp_rank.handle(parameterless=[split_msg()])
# async def _(state: T_State):
#     if "error" in state:
#         await UniMessage.text(state["error"]).send(reply_to=True)
#     mode = state["mode"]
#     group_id = event.group_id
#     group_member = await bot.get_group_member_list(group_id=group_id)
#     user_id_ls = [i["user_id"] for i in group_member]
#     binded_id = await UserData.filter(user_id__in=user_id_ls).values_list(
#         "osu_id", flat=True
#     )
#     info_ls = (
#         await InfoData.filter(osu_id__in=binded_id)
#         .filter(osu_mode=mode)
#         .filter(date=datetime.date.today())
#         .order_by("-pp")
#         .all()
#     )
#     icon_ls = [f"https://a.ppy.sh/{info.osu_id}" for info in info_ls]
#     tasks = [get_projectimg(i) for i in icon_ls]
#     icon_ls = await asyncio.gather(*tasks)
#     draw_len = len([i for i in info_ls if i.pp >= 100])
#     img = Image.new("RGBA", (1200, 85 + 82 * draw_len), (35, 42, 34, 255))
#     draw = ImageDraw.Draw(img)
#     draw.text(
#         (40, 10), f"{NGM[mode]}模式群内排名", font=Torus_Regular_25, fill=(255, 255, 255, 255)
#     )
#     draw.text((880, 10), "pp", font=Torus_Regular_25, fill=(255, 255, 255, 255))
#     draw.text((960, 10), "全球排名", font=Torus_Regular_25, fill=(255, 255, 255, 255))
#     draw.text((1070, 10), "国内排名", font=Torus_Regular_25, fill=(255, 255, 255, 255))
#     for index, (info, icon) in enumerate(zip(info_ls, icon_ls)):
#         if info.pp < 100:
#             continue
#         draw.rounded_rectangle(
#             (20, 55 + 82 * index, 1180, 45 + 82 * (index + 1)),
#             radius=10,
#             fill=(58, 70, 57, 255),
#         )
#         icon_img = Image.open(icon).convert("RGBA").resize((63, 63))
#         icon_img = draw_fillet(icon_img, 10)
#         img.alpha_composite(icon_img, (100, 60 + 82 * index))
#         for user in group_member:
#             if await UserData.filter(
#                 user_id=user["user_id"], osu_id=info.osu_id
#             ).first():
#                 name = user["card"] or user.get("nickname", "")
#                 break
#         else:
#             raise Exception("这不可能发生的")
#         draw.text(
#             (43, 70 + 82 * index),
#             f"#{index + 1}",
#             font=Torus_Regular_25,
#             fill=(255, 255, 255, 255),
#         )
#         draw.text(
#             (180, 70 + 82 * index),
#             name,
#             font=Torus_Regular_25,
#             fill=(166, 199, 163, 255),
#         )
#         draw.text(
#             (850, 70 + 82 * index),
#             f"{int(info.pp)}pp",
#             font=Torus_Regular_25,
#             fill=(255, 255, 255, 255),
#         )
#         draw.text(
#             (980, 70 + 82 * index),
#             f"{info.g_rank}",
#             font=Torus_Regular_25,
#             fill=(255, 255, 255, 255),
#         )
#         draw.text(
#             (1090, 70 + 82 * index),
#             f"{info.c_rank}",
#             font=Torus_Regular_25,
#             fill=(255, 255, 255, 255),
#         )
#
#     byt = BytesIO()
#     img.save(byt, "png")
#     await UniMessage.image(raw=byt).send(reply_to=True)



from datetime import datetime, timedelta
from io import BytesIO
from typing import Union

from PIL import ImageFilter, ImageEnhance, ImageDraw

from ..api import osu_api, sayo_api
from ..schema import SayoBeatmap
from ..file import get_projectimg
from .static import *
from .utils import crop_bg, calc_songlen, stars_diff, image2bytesio
from ..utils import GM


async def draw_bmap_info(mapid, op: bool = False) -> Union[str, BytesIO]:
    if op:
        info = await osu_api("map", map_id=mapid)
        if not info:
            return "未查询到地图"
        elif isinstance(info, str):
            return info
        mapid = info["beatmapset_id"]
    info = await sayo_api(mapid)
    if isinstance(info, str):
        return info
    sayo_info = SayoBeatmap(**info)
    if sayo_info.status == -1:
        return "在sayobot未查询到该地图"
    data = sayo_info.data

    coverurl = f"https://assets.ppy.sh/beatmaps/{mapid}/covers/cover@2x.jpg"
    cover = await get_projectimg(coverurl)
    # 新建
    if len(data.bid_data) > 20:
        im_h = 400 + 102 * 20
    else:
        im_h = 400 + 102 * (len(data.bid_data) - 1)
    im = Image.new("RGBA", (1200, im_h), (31, 41, 46, 255))
    draw = ImageDraw.Draw(im)
    # 背景
    cover_crop = await crop_bg("MP", cover)
    cover_gb = cover_crop.filter(ImageFilter.GaussianBlur(1))
    cover_img = ImageEnhance.Brightness(cover_gb).enhance(2 / 4.0)
    im.alpha_composite(cover_img, (0, 0))
    # 曲名
    draw.text((25, 15), data.title, font=Torus_SemiBold_40, anchor="lt")
    # 曲师
    draw.text((25, 70), f"by {data.artist}", font=Torus_SemiBold_20, anchor="lt")
    # mapper
    draw.text((25, 105), f"谱面作者: {data.creator}", font=Torus_SemiBold_20, anchor="lt")
    # rank时间
    if data.approved_date == -1:
        approved_date = "谱面状态可能非ranked"
    else:
        datearray = datetime.utcfromtimestamp(data.approved_date)
        approved_date = (datearray + timedelta(hours=8)).strftime("%Y-%m-%d %H:%M:%S")
    draw.text((25, 140), f"上架时间: {approved_date}", font=Torus_SemiBold_20, anchor="lt")
    # 来源
    draw.text((25, 175), f"Source: {data.source}", font=Torus_SemiBold_20, anchor="lt")
    # bpm
    draw.text((1150, 105), f"BPM: {data.bpm}", font=Torus_SemiBold_20, anchor="rt")
    # 曲长
    music_len = calc_songlen(data.bid_data[0].length)
    draw.text((1150, 140), f"length: {music_len}", font=Torus_SemiBold_20, anchor="rt")
    # Setid
    draw.text((1150, 35), f"Setid: {mapid}", font=Torus_SemiBold_20, anchor="rt")
    gmap = sorted(data.bid_data, key=lambda k: k.star, reverse=False)
    for num, cmap in enumerate(gmap):
        if num < 20:
            h_num = 102 * num
            # 难度
            draw.text((20, 320 + h_num), IconLs[GM[cmap.mode]], font=extra_30, anchor="lt")
            # 星星
            stars_bg = stars_diff(cmap.star)
            stars_img = stars_bg.resize((80, 30))
            im.alpha_composite(stars_img, (60, 320 + h_num))
            # diff
            im.alpha_composite(BarImg, (10, 365 + h_num))
            gc = ["CS", "HP", "OD", "AR"]
            for index, i in enumerate((cmap.CS, cmap.HP, cmap.OD, cmap.AR)):
                diff_len = int(200 * i / 10) if i <= 10 else 200
                diff_bg = Image.new("RGBA", (diff_len, 12), (255, 255, 255, 255))
                im.alpha_composite(diff_bg, (50 + 300 * index, 365 + h_num))
                draw.text(
                    (20 + 300 * index, 369 + h_num),
                    gc[index],
                    font=Torus_SemiBold_20,
                    anchor="lm",
                )
                draw.text(
                    (265 + 300 * index, 369 + h_num),
                    f"{i}",
                    font=Torus_SemiBold_20,
                    anchor="lm",
                    fill=(255, 204, 34, 255),
                )
                if index != 3:
                    draw.text(
                        (300 + 300 * index, 369 + h_num),
                        "|",
                        font=Torus_SemiBold_20,
                        anchor="lm",
                    )
            # 难度
            if cmap.star < 6.5:
                color = (0, 0, 0, 255)
            else:
                color = (255, 217, 102, 255)
            draw.text(
                (65, 335 + h_num),
                f"★{cmap.star:.2f}",
                font=Torus_SemiBold_20,
                anchor="lm",
                fill=color
            )
            # version
            draw.text(
                (150, 335 + h_num),
                f"{cmap.version}",
                font=Torus_SemiBold_20,
                anchor="lm",
            )
            # mapid
            draw.text(
                (1150, 328 + h_num),
                f"Mapid: {cmap.bid}",
                font=Torus_SemiBold_20,
                anchor="rm",
            )
            # maxcb
            draw.text(
                (700, 328 + h_num),
                f"Max Combo: {cmap.maxcombo}",
                font=Torus_SemiBold_20,
                anchor="lm",
            )
            # 分割线
            div = Image.new("RGBA", (1150, 2), (46, 53, 56, 255)).convert("RGBA")
            im.alpha_composite(div, (25, 400 + h_num))
        else:
            plusnum = f"+ {num - 19}"
            draw.text(
                (600, 350 + 102 * 20), plusnum, font=Torus_SemiBold_50, anchor="mm"
            )
    base = image2bytesio(im)
    im.close()
    return base



from nonebot import on_regex
from nonebot_plugin_alconna import UniMessage
from ..draw import draw_map_info
from nonebot.typing import T_State
from nonebot.params import RegexGroup

url_match = on_regex("https://osu.ppy.sh/beatmapsets/([0-9]+)#([^/]+/[0-9]+)")
url_1 = "https://osu.direct/api/d/"
url_2 = "https://txy1.sayobot.cn/beatmaps/download/novideo/"


@url_match.handle()
async def _url(state: T_State, bid: tuple = RegexGroup()):
    beatmap_id = bid[1].split("/")[1]
    url_total = f"kitsu镜像站：{url_1}{bid[0]}\n小夜镜像站：{url_2}{bid[0]}"
    state["para"] = beatmap_id
    state["mods"] = ""
    m = await draw_map_info(beatmap_id, state["mods"])
    await (UniMessage.image(raw=m) + "\n" + url_total).finish(reply_to=True)



import asyncio
from datetime import datetime, timedelta
from io import BytesIO
from typing import Optional, Union
from PIL import ImageFilter, ImageEnhance, ImageDraw, ImageSequence

from ..api import osu_api, get_map_bg, get_beatmap_attribute, get_sayo_map_info
from ..beatmap_stats_moder import with_mods
from ..schema import Score, Beatmap, User, BeatmapDifficultyAttributes
from ..mods import get_mods_list
from ..file import re_map, download_osu, user_cache_path, map_path
from ..utils import GMN, FGM
from ..pp import cal_pp, get_if_pp_ss_pp, get_ss_pp

from .static import *
from .utils import (
    draw_acc,
    draw_fillet,
    crop_bg,
    calc_songlen,
    stars_diff,
    get_modeimage,
    open_user_icon, is_close,
)


async def draw_score(
    project: str,
    uid: int,
    mode: str,
    mods: Optional[list[str]],
    best: int = 0,
    mapid: int = 0,
    is_name: bool = False,
) -> Union[str, BytesIO]:
    task0 = asyncio.create_task(osu_api(project, uid, mode, mapid, is_name=is_name))
    task1 = asyncio.create_task(osu_api("info", uid, mode, is_name=is_name))
    score_json = await task0
    if not score_json:
        return f"未查询到在 {GMN[mode]} 的游玩记录"
    elif isinstance(score_json, str):
        return score_json
    if project in ("recent", "pr"):
        if len(score_json) < best:
            return f"未查询到24小时内在 {GMN[mode]} 中第{best + 1}个游玩记录"
        score_info = Score(**score_json[best])
        grank = "--"
    elif project == "bp":
        score_ls = [Score(**i) for i in score_json]
        mods_ls = get_mods_list(score_ls, mods)
        if len(mods_ls) < best:
            return f'未找到开启 {"|".join(mods)} Mods的第{best}个成绩'
        score_info = Score(**score_json[mods_ls[best - 1]])
        grank = "--"
    else:
        raise "Project Error"
    # 从官网获取信息
    path = map_path / str(score_info.beatmap.beatmapset_id)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    osu = path / f"{score_info.beatmap.id}.osu"
    task = asyncio.create_task(get_beatmap_attribute(score_info.beatmap.id, mode))
    task2 = asyncio.create_task(osu_api("map", map_id=score_info.beatmap.id))
    if not osu.exists():
        await download_osu(score_info.beatmap.beatmapset_id, score_info.beatmap.id)
    info_json = await task1
    info = User(**info_json)
    user_path = user_cache_path / str(info.id)
    if not user_path.exists():
        user_path.mkdir(parents=True, exist_ok=True)
    map_json = await task2
    map_attribute_json = await task
    return await draw_score_pic(
        score_info,
        info,
        map_json,
        map_attribute_json,
        score_info.beatmap.id,
        score_info.beatmap.beatmapset_id,
    )


async def get_score_data(
    uid: int,
    mode: str,
    mods: Optional[list[str]],
    mapid: int = 0,
    is_name: bool = False,
):
    task = asyncio.create_task(get_beatmap_attribute(mapid, mode))
    task0 = asyncio.create_task(osu_api("score", uid, mode, mapid, is_name=is_name))
    task1 = asyncio.create_task(osu_api("info", uid, mode, is_name=is_name))
    task2 = asyncio.create_task(osu_api("map", map_id=mapid))
    task3 = asyncio.create_task(get_sayo_map_info(mapid, 1))
    score_json = await task0
    if not score_json:
        return f"未查询到在 {GMN[mode]} 的游玩记录"
    elif isinstance(score_json, str):
        return score_json
    if "score" not in score_json:
        score_ls = [Score(**i) for i in score_json["scores"]]
        if not score_ls:
            return f"未查询到在 {GMN[mode]} 的游玩记录"
        if mods:
            for score in score_ls:
                if mods == "NM" and not score.mods:
                    score_info = score
                    break
                if score.mods == mods:
                    score_info = score
                    break
            else:
                for score in score_ls:
                    if set(mods).issubset(set(score.mods)):
                        score_info = score
                        break
                else:
                    return f'未找到开启 {"|".join(mods)} Mods的成绩'
        else:
            score_info = score_ls[0]
    else:
        score_info = Score(**score_json["score"])
    sayo_map_info = await task3
    path = map_path / str(sayo_map_info.data.sid)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    osu = path / f"{mapid}.osu"
    if not osu.exists():
        await download_osu(sayo_map_info.data.sid, mapid)
    info_json = await task1
    info = User(**info_json)
    user_path = user_cache_path / str(info.id)
    if not user_path.exists():
        user_path.mkdir(parents=True, exist_ok=True)
    map_json = await task2
    map_attribute_json = await task
    return await draw_score_pic(
        score_info, info, map_json, map_attribute_json, mapid, sayo_map_info.data.sid
    )


async def draw_score_pic(
    score_info, info, map_json, map_attribute_json, bid, sid
) -> BytesIO:
    mapinfo = Beatmap(**map_json)
    original_mapinfo = mapinfo.copy()
    mapinfo = with_mods(mapinfo, score_info, score_info.mods)
    path = map_path / str(sid)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    # pp
    osu = path / f"{bid}.osu"
    pp_info = cal_pp(score_info, str(osu.absolute()))
    original_ss_pp_info = get_ss_pp(str(osu.absolute()), 0)
    if_pp, ss_pp = get_if_pp_ss_pp(score_info, str(osu.absolute()))
    # 新建图片
    im = Image.new("RGBA", (1500, 720))
    draw = ImageDraw.Draw(im)
    # 获取cover并裁剪，高斯，降低亮度
    cover = re_map(osu)
    cover_path = path / cover
    if not cover_path.exists():
        if bg := await get_map_bg(bid, sid, cover):
            with open(cover_path, "wb") as f:
                f.write(bg.getvalue())
    cover_crop = await crop_bg("BG", cover_path)
    cover_gb = cover_crop.filter(ImageFilter.GaussianBlur(3))
    cover_img = ImageEnhance.Brightness(cover_gb).enhance(2 / 4.0)
    im.alpha_composite(cover_img, (0, 0))
    # 获取成绩背景做底图
    bg = get_modeimage(FGM[score_info.mode])
    recent_bg = Image.open(bg).convert("RGBA")
    im.alpha_composite(recent_bg)
    # 模式
    draw.text((75, 75), IconLs[score_info.mode], font=extra_30, anchor="lt")
    # 难度星星
    stars_bg = stars_diff(pp_info.difficulty.stars)
    stars_img = stars_bg.resize((85, 37))
    im.alpha_composite(stars_img, (122, 72))
    if pp_info.difficulty.stars < 6.5:
        color = (0, 0, 0, 255)
    else:
        color = (255, 217, 102, 255)
    # 星级
    draw.text(
        (128, 90),
        f"★{pp_info.difficulty.stars:.2f}",
        font=Torus_SemiBold_20,
        anchor="lm",
        fill=color,
    )
    # mods
    if "HD" in score_info.mods or "FL" in score_info.mods:
        ranking = ["XH", "SH", "A", "B", "C", "D", "F"]
    else:
        ranking = ["X", "S", "A", "B", "C", "D", "F"]
    if score_info.mods:
        for mods_num, s_mods in enumerate(score_info.mods):
            mods_bg = osufile / "mods" / f"{s_mods}.png"
            mods_img = Image.open(mods_bg).convert("RGBA")
            im.alpha_composite(mods_img, (500 + 50 * mods_num, 160))
    # 成绩S-F
    rank_ok = False
    for rank_num, i in enumerate(ranking):
        rank_img = osufile / "ranking" / f"ranking-{i}.png"
        if rank_ok:
            rank_b = Image.open(rank_img).convert("RGBA").resize((48, 24))
            rank_new = Image.new("RGBA", rank_b.size, (0, 0, 0, 0))
            rank_bg = Image.blend(rank_new, rank_b, 0.5)
        elif i != score_info.rank:
            rank_b = Image.open(rank_img).convert("RGBA").resize((48, 24))
            rank_new = Image.new("RGBA", rank_b.size, (0, 0, 0, 0))
            rank_bg = Image.blend(rank_new, rank_b, 0.2)
        else:
            rank_bg = Image.open(rank_img).convert("RGBA").resize((48, 24))
            rank_ok = True
        im.alpha_composite(rank_bg, (75, 163 + 39 * rank_num))
    # 成绩+acc
    im = draw_acc(im, score_info.accuracy, score_info.mode)
    # 地区
    country = osufile / "flags" / f"{info.country_code}.png"
    country_bg = Image.open(country).convert("RGBA").resize((66, 45))
    im.alpha_composite(country_bg, (250, 577))
    # supporter
    if info.is_supporter:
        im.alpha_composite(SupporterBg.resize((40, 40)), (250, 640))
    map_attribute = BeatmapDifficultyAttributes(**map_attribute_json["attributes"])
    # 处理mania转谱cs
    if score_info.mode == "mania" and mapinfo.mode == "osu":
        temp_accuracy = mapinfo.accuracy
        convert = (mapinfo.count_sliders + mapinfo.count_spinners) / (
            mapinfo.count_circles + mapinfo.count_sliders + mapinfo.count_spinners
        )
        convert_od = round(temp_accuracy)
        if (
            convert < 0.20
            or (convert < 0.30 or round(mapinfo.cs) >= 5)
            and convert_od > 5
        ):
            mapinfo.cs = 7.0
        elif (convert < 0.30 or round(mapinfo.cs) >= 5) and convert_od <= 5:
            mapinfo.cs = 6.0
        elif convert > 0.60 and convert_od > 4:
            mapinfo.cs = 5.0
        elif convert > 0.60 and convert_od <= 4:
            mapinfo.cs = 4.0
        else:
            temp_accuracy += 1
            mapinfo.cs = max(4.0, min(temp_accuracy, 7.0))
    # cs, ar, od, hp
    mapdiff = [mapinfo.cs, mapinfo.drain, mapinfo.accuracy, mapinfo.ar]
    original_mapdiff = [
        original_mapinfo.cs,
        original_mapinfo.drain,
        original_mapinfo.accuracy,
        original_mapinfo.ar,
    ]

    for num, (orig, new) in enumerate(zip(original_mapdiff, mapdiff)):
        orig_difflen = int(250 * max(0, orig) / 10) if orig <= 10 else 250
        new_difflen = int(250 * max(0, new) / 10) if new <= 10 else 250
        if is_close(new, orig):
            color = (255, 255, 255, 255)
            diff_len = Image.new("RGBA", (orig_difflen, 8), color)
            im.alpha_composite(diff_len, (1190, 306 + 35 * num))
        elif new > orig and not (score_info.mode == "mania" and mapinfo.mode == "osu"):
            color = (198, 92, 102, 255)
            orig_color = (246, 136, 144, 255)
            new_diff_len = Image.new("RGBA", (new_difflen, 8), color)
            im.alpha_composite(new_diff_len, (1190, 306 + 35 * num))
            orig_diff_len = Image.new("RGBA", (orig_difflen, 8), orig_color)
            im.alpha_composite(orig_diff_len, (1190, 306 + 35 * num))
        elif new < orig and not (score_info.mode == "mania" and mapinfo.mode == "osu"):
            color = (161, 212, 238, 255)
            orig_color = (255, 255, 255, 255)
            orig_diff_len = Image.new("RGBA", (orig_difflen, 8), orig_color)
            im.alpha_composite(orig_diff_len, (1190, 306 + 35 * num))
            new_diff_len = Image.new("RGBA", (new_difflen, 8), color)
            im.alpha_composite(new_diff_len, (1190, 306 + 35 * num))
        else:
            raise Exception('没有这种情况')
        if new == round(new):
            draw.text(
                (1470, 310 + 35 * num),
                "%.0f" % new,
                font=Torus_SemiBold_20,
                anchor="mm",
            )
        else:
            draw.text(
                (1470, 310 + 35 * num),
                "%.2f" % new if new != round(new, 1) else "%.1f" % new,
                font=Torus_SemiBold_20,
                anchor="mm",
            )
    # stardiff
    stars = pp_info.difficulty.stars
    original_stars = original_ss_pp_info.difficulty.stars
    if stars > original_stars:
        color = (198, 92, 102, 255)
        orig_color = (246, 111, 34, 255)
        new_difflen = int(250 * max(0.0, stars) / 10) if stars <= 10 else 250
        new_diff_len = Image.new("RGBA", (new_difflen, 8), color)
        im.alpha_composite(new_diff_len, (1190, 446))
        orig_difflen = (
            int(250 * max(0.0, original_stars) / 10) if original_stars <= 10 else 250
        )
        orig_diff_len = Image.new("RGBA", (orig_difflen, 8), orig_color)
        im.alpha_composite(orig_diff_len, (1190, 446))
    elif stars < original_stars:
        color = (161, 187, 127, 255)
        orig_color = (255, 204, 34, 255)
        orig_difflen = (
            int(250 * max(0.0, original_stars) / 10) if original_stars <= 10 else 250
        )
        orig_diff_len = Image.new("RGBA", (orig_difflen, 8), orig_color)
        im.alpha_composite(orig_diff_len, (1190, 446))
        new_difflen = int(250 * max(0.0, stars) / 10) if stars <= 10 else 250
        new_diff_len = Image.new("RGBA", (new_difflen, 8), color)
        im.alpha_composite(new_diff_len, (1190, 446))
    else:
        color = (255, 204, 34, 255)
        difflen = int(250 * stars / 10) if stars <= 10 else 250
        diff_len = Image.new("RGBA", (difflen, 8), color)
        im.alpha_composite(diff_len, (1190, 446))
    draw.text((1470, 450), f"{stars:.2f}", font=Torus_SemiBold_20, anchor="mm")
    # 时长 - 滑条
    diff_info = (
        calc_songlen(mapinfo.total_length),
        f"{mapinfo.bpm:.1f}",
        mapinfo.count_circles,
        mapinfo.count_sliders,
    )
    for num, i in enumerate(diff_info):
        draw.text(
            (1070 + 120 * num, 245),
            f"{i}",
            font=Torus_Regular_20,
            anchor="lm",
            fill=(255, 204, 34, 255),
        )
    # 状态
    draw.text(
        (1400, 184), mapinfo.status.capitalize(), font=Torus_SemiBold_20, anchor="mm"
    )
    # mapid
    draw.text((1485, 90), f"Mapid: {bid}", font=Torus_SemiBold_20, anchor="rm")
    # 曲名
    draw.text(
        (75, 38),
        f"{mapinfo.beatmapset.title} | by {mapinfo.beatmapset.artist_unicode}",
        font=Torus_SemiBold_30,
        anchor="lm",
    )
    # 谱面版本，mapper
    draw.text(
        (225, 90),
        f"{mapinfo.version} | 谱师: {mapinfo.beatmapset.creator}",
        font=Torus_SemiBold_20,
        anchor="lm",
    )
    # 评价
    draw.text((316, 307), score_info.rank, font=Venera_75, anchor="mm")
    # 分数
    draw.text((498, 251), f"{score_info.score:,}", font=Torus_Regular_75, anchor="lm")
    # 时间
    draw.text((498, 341), "达成时间:", font=Torus_SemiBold_20, anchor="lm")
    old_time = datetime.strptime(
        score_info.created_at.replace("Z", ""), "%Y-%m-%dT%H:%M:%S"
    )
    new_time = (old_time + timedelta(hours=8)).strftime("%Y-%m-%d %H:%M:%S")
    draw.text((630, 341), new_time, font=Torus_SemiBold_20, anchor="lm")
    # 全球排名
    # draw.text((583, 410), grank, font=Torus_SemiBold_25, anchor='mm')
    # 左下玩家名
    draw.text((250, 530), info.username, font=Torus_SemiBold_30, anchor="lm")
    # 国内排名
    draw.text(
        (325, 610),
        f"#{info.statistics.country_rank}",
        font=Torus_SemiBold_25,
        anchor="lm",
    )
    if score_info.mode == "osu":
        draw.text((720, 550), ss_pp, font=Torus_Regular_30, anchor="mm")
        draw.text((840, 550), if_pp, font=Torus_Regular_30, anchor="mm")
        draw.text((960, 550), f"{pp_info.pp:.0f}", font=Torus_Regular_30, anchor="mm")
        draw.text(
            (720, 645), f"{pp_info.pp_aim:.0f}", font=Torus_Regular_30, anchor="mm"
        )
        draw.text(
            (840, 645), f"{pp_info.pp_speed:.0f}", font=Torus_Regular_30, anchor="mm"
        )
        draw.text(
            (960, 645), f"{pp_info.pp_acc:.0f}", font=Torus_Regular_30, anchor="mm"
        )
        draw.text(
            (1157, 550),
            f"{score_info.accuracy * 100:.2f}%",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1385, 550),
            f"{score_info.max_combo:,}/{map_attribute.max_combo:,}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1100, 645),
            f"{score_info.statistics.count_300}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1214, 645),
            f"{score_info.statistics.count_100}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1328, 645),
            f"{score_info.statistics.count_50}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1442, 645),
            f"{score_info.statistics.count_miss}",
            font=Torus_Regular_30,
            anchor="mm",
        )
    elif score_info.mode == "taiko":
        draw.text(
            (1118, 550),
            f"{score_info.accuracy * 100:.2f}%",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1270, 550), f"{score_info.max_combo:,}", font=Torus_Regular_30, anchor="mm"
        )
        draw.text(
            (1420, 550), f"{pp_info.pp:.0f}/{ss_pp}", font=Torus_Regular_30, anchor="mm"
        )
        draw.text(
            (1118, 645),
            f"{score_info.statistics.count_300}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1270, 645),
            f"{score_info.statistics.count_100}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1420, 645),
            f"{score_info.statistics.count_miss}",
            font=Torus_Regular_30,
            anchor="mm",
        )
    elif score_info.mode == "fruits":
        draw.text(
            (1083, 550),
            f"{score_info.accuracy * 100:.2f}%",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1247, 550),
            f"{score_info.max_combo}/{map_attribute.max_combo}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1411, 550), f"{pp_info.pp:.0f}/{ss_pp}", font=Torus_Regular_30, anchor="mm"
        )
        draw.text(
            (1062, 645),
            f"{score_info.statistics.count_300}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1185, 645),
            f"{score_info.statistics.count_100}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1309, 645),
            f"{score_info.statistics.count_katu}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1432, 645),
            f"{score_info.statistics.count_miss}",
            font=Torus_Regular_30,
            anchor="mm",
        )
    else:
        draw.text(
            (1002, 580),
            f"{score_info.statistics.count_geki / score_info.statistics.count_300 :.1f}:1"
            if score_info.statistics.count_300 != 0
            else "∞:1",
            font=Torus_Regular_20,
            anchor="mm",
        )
        draw.text(
            (1002, 550),
            f"{score_info.accuracy * 100:.2f}%",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1197, 550), f"{score_info.max_combo}", font=Torus_Regular_30, anchor="mm"
        )
        draw.text(
            (1395, 550), f"{pp_info.pp:.0f}/{ss_pp}", font=Torus_Regular_30, anchor="mm"
        )
        draw.text(
            (953, 645),
            f"{score_info.statistics.count_geki}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1051, 645),
            f"{score_info.statistics.count_300}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1150, 645),
            f"{score_info.statistics.count_katu}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1249, 645),
            f"{score_info.statistics.count_100}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1347, 645),
            f"{score_info.statistics.count_50}",
            font=Torus_Regular_30,
            anchor="mm",
        )
        draw.text(
            (1445, 645),
            f"{score_info.statistics.count_miss}",
            font=Torus_Regular_30,
            anchor="mm",
        )
    user_icon = await open_user_icon(info)
    gif_frames = []
    if not getattr(user_icon, "is_animated", False):
        icon_bg = user_icon.convert("RGBA").resize((170, 170))
        icon_img = draw_fillet(icon_bg, 15)
        im.alpha_composite(icon_img, (60, 510))
        byt = BytesIO()
        im.save(byt, "png")
        im.close()
        user_icon.close()
        return byt
    for gif_frame in ImageSequence.Iterator(user_icon):
        # 将 GIF 图片中的每一帧转换为 RGBA 模式
        gif_frame = gif_frame.convert("RGBA").resize((170, 170))
        gif_frame = draw_fillet(gif_frame, 15)
        # 创建一个新的 RGBA 图片，将 PNG 图片作为背景，将当前帧添加到背景上
        rgba_frame = Image.new("RGBA", im.size, (0, 0, 0, 0))
        rgba_frame.paste(im, (0, 0), im)
        rgba_frame.paste(gif_frame, (60, 510), gif_frame)
        # 将 RGBA 图片转换为 RGB 模式，并添加到 GIF 图片中
        gif_frames.append(rgba_frame)
    gif_bytes = BytesIO()
    # 保存 GIF 图片
    gif_frames[0].save(
        gif_bytes, format="gif", save_all=True, append_images=gif_frames[1:]
    )
    # 输出
    gif_frames[0].close()
    user_icon.close()
    return gif_bytes



from nonebot import on_command
from nonebot.internal.adapter import Message
from nonebot.params import CommandArg
from nonebot_plugin_alconna import UniMessage
from ..info import get_bg

getbg = on_command("getbg", priority=11, block=True)


@getbg.handle()
async def _get_bg(bg: Message = CommandArg()):
    bg = bg.extract_plain_text().strip()
    if not bg:
        await UniMessage.text("请输入需要提取BG的地图ID").finish(reply_to=True)
    byt = await get_bg(bg)
    if isinstance(byt, str):
        await UniMessage.text(byt).finish(reply_to=True)
    await UniMessage.image(raw=byt).finish(reply_to=True)



from arclet.alconna import Alconna, Args, CommandMeta
from nonebot import on_command, get_driver
from nonebot.internal.adapter import Event
from nonebot_plugin_alconna import (
    on_alconna,
    UniMessage,
    AlconnaMatch,
    image_fetch,
    Match,
    Target,
    SupportScope,
)
from nonebot_plugin_alconna.uniseg import Image
from nonebot.typing import T_State

from ..database import UserData
from ..file import user_cache_path, save_info_pic


from .utils import split_msg

update_pic = on_alconna(
    Alconna(
        "更新背景",
        Args["img?", Image],
        meta=CommandMeta(example="/更新背景 [图片]"),
    ),
    skip_for_unmatch=False,
    use_cmd_start=True,
    aliases=("更改背景",),
)
update_info = on_command("update", priority=11, block=True, aliases={"更新信息"})
clear_background = on_command("清空背景", priority=11, block=True, aliases={"清除背景", "重置背景"})


@update_pic.handle()
async def _(
    event: Event, state: T_State, img: Match[bytes] = AlconnaMatch("img", image_fetch)
):
    qq = event.get_user_id()
    user_data = await UserData.get_or_none(user_id=int(qq))
    state["user"] = user_data.osu_id if user_data else 0
    if state["user"] == 0:
        await UniMessage.text("该账号尚未绑定，请输入 /bind 用户名 绑定账号").finish(reply_to=True)
    if not img.available:
        await UniMessage.text("请在指令之后附上图片").finish(reply_to=True)
    user = state["user"]
    pic_url = img.result
    await save_info_pic(str(user), pic_url)
    msg = f"收到自{qq}的更新背景申请" + UniMessage.image(raw=pic_url)
    for superuser in get_driver().config.superusers:
        await Target.user(superuser, SupportScope.qq_client).send(msg)
    await UniMessage.text("更新背景成功").finish(reply_to=True)


@update_info.handle(parameterless=[split_msg()])
async def _(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).send(reply_to=True)
        return
    user = state["user"]
    path = user_cache_path / str(user) / "icon.png"
    gif_path = user_cache_path / str(user) / "icon.gif"
    if path.exists():
        path.unlink()
    if gif_path.exists():
        gif_path.unlink()
    await UniMessage.text("个人信息更新成功").send(reply_to=True)


@clear_background.handle(parameterless=[split_msg()])
async def _(state: T_State):
    if "error" in state:
        await UniMessage.text(state["error"]).send(reply_to=True)
        return
    user = state["user"]
    path = user_cache_path / str(user) / "info.png"
    if path.exists():
        path.unlink()
        await UniMessage.text("背景图片清除成功").send(reply_to=True)
        return
    else:
        await UniMessage.text("您还没有设置背景或已成功清除背景").send(reply_to=True)



import asyncio
from nonebot import on_command
from nonebot.internal.adapter import Event, Message
from nonebot.params import CommandArg
from nonebot_plugin_alconna import UniMessage

from ..info import bind_user_info
from ..database import UserData


bind = on_command("bind", priority=11, block=True)
unbind = on_command("unbind", priority=11, block=True)
lock = asyncio.Lock()


@bind.handle()
async def _bind(event: Event, name: Message = CommandArg()):
    name = name.extract_plain_text().strip()
    if not name:
        await UniMessage.text("请在指令后输入您的 osuid").finish(reply_to=True)
    async with lock:
        if user := await UserData.get_or_none(user_id=event.get_user_id()):
            await UniMessage.text(f"您已绑定{user.osu_name}，如需要解绑请输入/unbind").finish(reply_to=True)
        msg = await bind_user_info("bind", name, event.get_user_id(), True)
    await UniMessage.text(msg).finish(reply_to=True)


@unbind.handle()
async def _unbind(event: Event):
    if _ := await UserData.get_or_none(user_id=event.get_user_id()):
        await UserData.filter(user_id=event.get_user_id()).delete()
        await UniMessage.text("解绑成功！").send(reply_to=True)
    else:
        await UniMessage.text("尚未绑定，无需解绑").send(reply_to=True)



from nonebot import on_command
from nonebot.internal.adapter import Message
from nonebot.params import CommandArg
from nonebot.typing import T_State
from nonebot_plugin_alconna import UniMessage

from ..mania import generate_preview_pic
from ..api import osu_api
from ..file import download_tmp_osu

generate_preview = on_command(
    "预览", aliases={"preview", "完整预览"}, priority=11, block=True
)


@generate_preview.handle()
async def _(
    state: T_State,
    args: Message = CommandArg(),
):
    osu_id = args.extract_plain_text().strip()
    if not osu_id or not osu_id.isdigit():
        await UniMessage.text("请输入正确的地图mapID").finish(reply_to=True)
    data = await osu_api("map", map_id=int(osu_id))
    if not data:
        await UniMessage.text("未查询到该地图").finish(reply_to=True)
    if isinstance(data, str):
        await UniMessage.text(data).finish(reply_to=True)
    osu = await download_tmp_osu(osu_id)
    if state["_prefix"]["command"][0] == "完整预览":
        pic = await generate_preview_pic(osu, True)
    else:
        pic = await generate_preview_pic(osu)
    await UniMessage.image(raw=pic).finish(reply_to=True)



GM = {0: "osu", 1: "taiko", 2: "fruits", 3: "mania"}
NGM = {"0": "osu", "1": "taiko", "2": "fruits", "3": "mania"}
GMN = {"osu": "Std", "taiko": "Taiko", "fruits": "Ctb", "mania": "Mania"}
FGM = {"osu": 0, "taiko": 1, "fruits": 2, "mania": 3}


def mods2list(args: str) -> list:
    args = args.replace(" ", "").replace(",", "").replace("，", "")
    args = args.upper()
    return [args[i : i + 2] for i in range(0, len(args), 2)]
